"""
=========================================================================
test_cases.py
=========================================================================
Centralized test case repository for the SystemVerilog backend.

Author : Peitian Pan
Date   : Dec 16, 2019
"""
from os.path import dirname
from textwrap import dedent

from pymtl3 import *
from pymtl3.passes.backends.verilog import *
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.testcases import (
    Bits32Foo,
    Bits32x5Foo,
    CaseArrayBits32IfcInUpblkComp,
    CaseBehavioralArraySubCompArrayStructIfcComp,
    CaseBits32ArrayConnectSubCompAttrComp,
    CaseBits32ArraySubCompAttrUpblkComp,
    CaseBits32BitSelUpblkComp,
    CaseBits32ConnectSubCompAttrComp,
    CaseBits32FooInBits32OutComp,
    CaseBits32FooNoArgBehavioralComp,
    CaseBits32FooToBits32Comp,
    CaseBits32SubCompAttrUpblkComp,
    CaseBits32ToBits32FooComp,
    CaseBits32x2ConcatComp,
    CaseBits32x2ConcatConstComp,
    CaseBits32x2ConcatFreeVarComp,
    CaseBits32x2ConcatMixedComp,
    CaseBits32x2ConcatUnpackedSignalComp,
    CaseBits64PartSelUpblkComp,
    CaseBits64SextInComp,
    CaseBits64TruncInComp,
    CaseBits64ZextInComp,
    CaseBitSelOverBitSelComp,
    CaseBitSelOverPartSelComp,
    CaseBoolTmpVarComp,
    CaseChildExplicitModuleName,
    CaseConnectArrayBits32Comp,
    CaseConnectArrayBits32FooIfcComp,
    CaseConnectArrayNestedIfcComp,
    CaseConnectArrayStructAttrToOutComp,
    CaseConnectArraySubCompArrayStructIfcComp,
    CaseConnectBitsConstToOutComp,
    CaseConnectBitSelToOutComp,
    CaseConnectConstStructAttrToOutComp,
    CaseConnectConstToOutComp,
    CaseConnectInToWireComp,
    CaseConnectLiteralStructComp,
    CaseConnectNestedStructPackedArrayComp,
    CaseConnectPassThroughLongNameComp,
    CaseConnectSliceToOutComp,
    CaseConnectValRdyIfcComp,
    CaseConnectValRdyIfcUpblkComp,
    CaseConstStructInstComp,
    CaseDefaultBitsComp,
    CaseElifBranchComp,
    CaseFixedSizeSliceComp,
    CaseForLoopEmptySequenceComp,
    CaseForRangeLowerUpperStepPassThroughComp,
    CaseHeteroCompArrayComp,
    CaseIfBasicComp,
    CaseIfBoolOpInForStmtComp,
    CaseIfDanglingElseInnerComp,
    CaseIfDanglingElseOutterComp,
    CaseIfExpBothImplicitComp,
    CaseIfExpInForStmtComp,
    CaseIfExpUnaryOpInForStmtComp,
    CaseIfTmpVarInForStmtComp,
    CaseInterfaceArrayNonStaticIndexComp,
    CaseIntToBits32FooComp,
    CaseLambdaConnectComp,
    CaseLambdaConnectWithListComp,
    CaseNestedIfComp,
    CaseNestedStructPackedArrayUpblkComp,
    CasePartSelOverBitSelComp,
    CasePartSelOverPartSelComp,
    CasePassThroughComp,
    CasePythonClassAttr,
    CaseReducesInx3OutComp,
    CaseSequentialPassThroughComp,
    CaseSizeCastPaddingStructPort,
    CaseStructPackedArrayUpblkComp,
    CaseStructUnique,
    CaseTmpVarInUpdateffComp,
    CaseTypeBundle,
    CaseVerilogReservedComp,
    NestedStructPackedPlusScalar,
    ThisIsABitStructWithSuperLongName,
    set_attributes,
)
from pymtl3.passes.testcases.test_cases import _check, _set

# Verilog test cases

class Bits32VRegComp( VerilogPlaceholder, Component ):
  def construct( s ):
    s.d = InPort( Bits32 )
    s.q = OutPort( Bits32 )
    s.set_metadata( VerilogPlaceholderPass.src_file, dirname(__file__)+'/VReg.v' )
    s.set_metadata( VerilogPlaceholderPass.top_module, 'VReg' )

class Bits32VRegPassThroughComp( VerilogPlaceholder, Component ):
  def construct( s ):
    s.d = InPort( Bits32 )
    s.q = OutPort( Bits32 )
    s.set_metadata( VerilogPlaceholderPass.src_file, dirname(__file__)+'/VRegPassThrough.v' )
    s.set_metadata( VerilogPlaceholderPass.top_module, 'VRegPassThrough' )
    s.set_metadata( VerilogPlaceholderPass.v_include, [dirname(__file__)] )

class CasePlaceholderTranslationVReg:
  DUT = Bits32VRegComp
  TV_IN = \
  _set( 'd', Bits32, 0 )
  TV_OUT = \
  _check( 'q', Bits32, 1 )
  TV = \
  [
      [  1,  0 ],
      [  2,  1 ],
      [ 42,  2 ],
      [ -1, 42 ],
      [  0, -1 ],
  ]

class CasePlaceholderTranslationRegIncr:
  class DUT( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.reg_ = Bits32VRegComp()
      s.reg_.d //= s.in_
      s.out //= lambda: s.reg_.q + Bits32(1)
  TV_IN = \
  _set( 'in_', Bits32, 0 )
  TV_OUT = \
  _check( 'out', Bits32, 1 )
  TV = \
  [
      [  1,  1 ],
      [  2,  2 ],
      [ 42,  3 ],
      [ -1, 43 ],
      [  0,  0 ],
  ]

class CaseVIncludePopulation:
  class DUT( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.reg_ = Bits32VRegPassThroughComp()
      s.reg_.d //= s.in_
      s.reg_.q //= s.out
  TV_IN = \
  _set( 'in_', Bits32, 0 )
  TV_OUT = \
  _check( 'out', Bits32, 1 )
  TV = \
  [
      [  1,  0 ],
      [  2,  1 ],
      [ 42,  2 ],
      [ -1, 42 ],
      [  0, -1 ],
  ]

class CaseVLibsTranslation:
  class DUT( VerilogPlaceholder, Component ):
    def construct( s ):
      s.d = InPort( Bits32 )
      s.q = OutPort( Bits32 )
      s.set_metadata( VerilogPlaceholderPass.src_file, dirname(__file__)+'/VRegPassThrough.v' )
      s.set_metadata( VerilogPlaceholderPass.top_module, 'VRegPassThrough' )
      s.set_metadata( VerilogPlaceholderPass.v_libs, [dirname(__file__)+'/VReg.v'] )
  TV_IN = \
  _set( 'd', Bits32, 0 )
  TV_OUT = \
  _check( 'q', Bits32, 1 )
  TV = \
  [
      [  1,  0 ],
      [  2,  1 ],
      [ 42,  2 ],
      [ -1, 42 ],
      [  0, -1 ],
  ]

class CaseMultiPlaceholderImport:
  class DUT( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.a_reg = Bits32VRegComp()
      s.b_reg = Bits32VRegComp()
      s.a_reg.d //= s.in_
      s.a_reg.q //= s.b_reg.d
      s.b_reg.q //= s.out
  TV_IN  = _set( 'in_', Bits32, 0 )
  TV_OUT = _check( 'out', Bits32, 1 )
  TV = \
  [
      [  1,  0 ],
      [  2,  0 ],
      [ 42,  1 ],
      [ -1,  2 ],
      [  0, 42 ],
      [  0, -1 ],
  ]

class CasePlaceholderTranslationRegIncr:
  class DUT( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.reg_ = Bits32VRegComp()
      s.reg_.d //= s.in_
      s.out //= lambda: s.reg_.q + Bits32(1)
  TV_IN = \
  _set( 'in_', Bits32, 0 )
  TV_OUT = \
  _check( 'out', Bits32, 1 )
  TV = \
  [
      [  1,  1 ],
      [  2,  2 ],
      [ 42,  3 ],
      [ -1, 43 ],
      [  0,  0 ],
  ]

CaseSizeCastPaddingStructPort = set_attributes( CaseSizeCastPaddingStructPort,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = 64'( in_ );
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo__foo_32;

        module DUT_noparam
        (
          input logic [0:0] clk,
          input Bits32Foo__foo_32 in_,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = 64'( in_ );
          end

        endmodule
    '''
)

CasePassThroughComp = set_attributes( CasePassThroughComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = in_;
          end

        endmodule
    '''
)

CaseSequentialPassThroughComp = set_attributes( CaseSequentialPassThroughComp,
    'REF_UPBLK',
    '''\
        always_ff @(posedge clk) begin : upblk
          out <= in_;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_ff @(posedge clk) begin : upblk
            out <= in_;
          end

        endmodule
    '''
)

CaseConnectPassThroughLongNameComp = set_attributes( CaseConnectPassThroughLongNameComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_,
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONN',
    '''\
        assign out = in_;
    ''',
    'REF_SRC',
    '''\
        module DUT__2ae3b6d12e9855d1
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = in_;

        endmodule
    ''',
    'REF_STRUCT',
    '''\
    ''',
)

CaseLambdaConnectComp = set_attributes( CaseLambdaConnectComp,
    'REF_UPBLK',
    '''\
        always_comb begin : _lambda__s_out
          out = in_ + 32'd42;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : _lambda__s_out
            out = in_ + 32'd42;
          end

        endmodule
    '''
)

CaseLambdaConnectWithListComp = set_attributes( CaseLambdaConnectWithListComp,
    'REF_UPBLK',
    '''\
        always_comb begin : _lambda__s_out_1_
          out[1'd1] = in_ + 32'd42;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out [0:1],
          input logic [0:0] reset
        );

          always_comb begin : _lambda__s_out_1_
            out[1'd1] = in_ + 32'd42;
          end

        endmodule
    '''
)

CaseBits32x2ConcatComp = set_attributes( CaseBits32x2ConcatComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_1, in_2 };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { in_1, in_2 };
          end

        endmodule
    '''
)

CaseBits32x2ConcatConstComp = set_attributes( CaseBits32x2ConcatConstComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { 32'd42, 32'd0 };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { 32'd42, 32'd0 };
          end

        endmodule
    '''
)

CaseBits32x2ConcatMixedComp = set_attributes( CaseBits32x2ConcatMixedComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_, 32'd0 };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { in_, 32'd0 };
          end

        endmodule
    '''
)

CaseBits64SextInComp = set_attributes( CaseBits64SextInComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { { 32 { in_[31] } }, in_ };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { { 32 { in_[31] } }, in_ };
          end

        endmodule
    '''
)

CaseBits64ZextInComp = set_attributes( CaseBits64ZextInComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { { 32 { 1'b0 } }, in_ };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          // @update
          // def upblk():
          //   s.out = zext( s.in_, 64 )

          always_comb begin : upblk
            out = { { 32 { 1'b0 } }, in_ };
          end

        endmodule
    '''
)

CaseBits64TruncInComp = set_attributes( CaseBits64TruncInComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = 8'(in_);
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [7:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = 8'(in_);
          end

        endmodule
    '''
)

CaseBits32x2ConcatFreeVarComp = set_attributes( CaseBits32x2ConcatFreeVarComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_, 1'( __const__STATE_IDLE ) };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [32:0] out,
          input logic [0:0] reset
        );
          localparam logic [0:0] __const__STATE_IDLE = 1'd0;

          always_comb begin : upblk
            out = { in_, 1'( __const__STATE_IDLE ) };
          end

        endmodule
    '''
)

CaseBits32x2ConcatUnpackedSignalComp = set_attributes( CaseBits32x2ConcatUnpackedSignalComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_[1'd0], in_[1'd1] };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:1],
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { in_[1'd0], in_[1'd1] };
          end

        endmodule
    '''
)

CaseBits32BitSelUpblkComp = set_attributes( CaseBits32BitSelUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_[5'd1];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [0:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = in_[5'd1];
          end

        endmodule
    '''
)

CaseBits64PartSelUpblkComp = set_attributes( CaseBits64PartSelUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_[6'd35:6'd4];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [63:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = in_[6'd35:6'd4];
          end

        endmodule
    '''
)

CaseStructUnique = set_attributes( CaseStructUnique,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          ST_A_wire = { 16'd1, 32'd2 };
          ST_B_wire = { 16'd3, 32'd4 };
          out = ST_A_wire.a_bar + ST_B_wire.b_bar;
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [15:0] a_foo;
          logic [31:0] a_bar;
        } ST__a_foo_16__a_bar_32;

        typedef struct packed {
          logic [15:0] b_foo;
          logic [31:0] b_bar;
        } ST__b_foo_16__b_bar_32;

        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          ST__a_foo_16__a_bar_32 ST_A_wire;
          ST__b_foo_16__b_bar_32 ST_B_wire;

          always_comb begin : upblk
            ST_A_wire = { 16'd1, 32'd2 };
            ST_B_wire = { 16'd3, 32'd4 };
            out = ST_A_wire.a_bar + ST_B_wire.b_bar;
          end

        endmodule
    '''
)

CasePythonClassAttr = set_attributes( CasePythonClassAttr,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out1 = 32'd42;
          out2 = 32'd1;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out1,
          output logic [31:0] out2,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out1 = 32'd42;
            out2 = 32'd1;
          end

        endmodule
    '''
)

CaseTypeBundle = set_attributes( CaseTypeBundle,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out1 = 32'd42;
          out2 = 32'd1;
          out3 = 32'd1;
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo__foo_32;

        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out1,
          output Bits32Foo__foo_32 out2,
          output Bits32Foo__foo_32 out3,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out1 = 32'd42;
            out2 = 32'd1;
            out3 = 32'd1;
          end

        endmodule
    '''
)

CaseTmpVarInUpdateffComp = set_attributes( CaseTmpVarInUpdateffComp,
    'REF_UPBLK',
    '''\
        always_ff @(posedge clk) begin : upblk
          if ( ~reset ) begin
            __tmpvar__upblk_val_next = 32'd42;
            out <= __tmpvar__upblk_val_next;
          end
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          logic [31:0] __tmpvar__upblk_val_next;

          always_ff @(posedge clk) begin : upblk
            if ( ~reset ) begin
              __tmpvar__upblk_val_next = 32'd42;
              out <= __tmpvar__upblk_val_next;
            end
          end

        endmodule
    '''
)

CaseBoolTmpVarComp = set_attributes( CaseBoolTmpVarComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          __tmpvar__upblk_matched = in_ == 32'd0;
          if ( __tmpvar__upblk_matched ) begin
            out = 32'd1;
          end
          else
            out = 32'd0;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          logic [0:0] __tmpvar__upblk_matched;

          always_comb begin : upblk
            __tmpvar__upblk_matched = in_ == 32'd0;
            if ( __tmpvar__upblk_matched ) begin
              out = 32'd1;
            end
            else
              out = 32'd0;
          end

        endmodule
    '''
)

CaseDefaultBitsComp = set_attributes( CaseDefaultBitsComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = 32'd0;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = 32'd0;
          end

        endmodule
    '''
)


CaseBits32FooToBits32Comp = set_attributes( CaseBits32FooToBits32Comp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_;
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo__foo_32;

        module DUT_noparam
        (
          input logic [0:0] clk,
          input Bits32Foo__foo_32 in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = in_;
          end

        endmodule
    '''
)

CaseBits32ToBits32FooComp = set_attributes( CaseBits32ToBits32FooComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_;
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo__foo_32;

        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output Bits32Foo__foo_32 out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = in_;
          end

        endmodule
    '''
)

CaseIntToBits32FooComp = set_attributes( CaseIntToBits32FooComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = 32'd42;
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo__foo_32;

        module DUT_noparam
        (
          input logic [0:0] clk,
          output Bits32Foo__foo_32 out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = 32'd42;
          end

        endmodule
    '''
)

CaseReducesInx3OutComp = set_attributes( CaseReducesInx3OutComp,
    'REF_UPBLK',
    '''\
        always_comb begin : v_reduce
          out = ( ( & in_1 ) & ( | in_2 ) ) | ( ^ in_3 );
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          input logic [31:0] in_3,
          output logic [0:0] out,
          input logic [0:0] reset
        );

          always_comb begin : v_reduce
            out = ( ( & in_1 ) & ( | in_2 ) ) | ( ^ in_3 );
          end

        endmodule
    '''
)

CaseIfBasicComp = set_attributes( CaseIfBasicComp,
    'REF_UPBLK',
    '''\
        always_comb begin : if_basic
          if ( in_[4'd7:4'd0] == 8'd255 ) begin
            out = in_[4'd15:4'd8];
          end
          else
            out = 8'd0;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [15:0] in_,
          output logic [7:0] out,
          input logic [0:0] reset
        );

          always_comb begin : if_basic
            if ( in_[4'd7:4'd0] == 8'd255 ) begin
              out = in_[4'd15:4'd8];
            end
            else
              out = 8'd0;
          end

        endmodule
    '''
)

CaseIfDanglingElseInnerComp = set_attributes( CaseIfDanglingElseInnerComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          if ( 1'd1 ) begin
            if ( 1'd0 ) begin
              out = in_1;
            end
            else
              out = in_2;
          end
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            if ( 1'd1 ) begin
              if ( 1'd0 ) begin
                out = in_1;
              end
              else
                out = in_2;
            end
          end

        endmodule
    '''
)

CaseIfDanglingElseOutterComp = set_attributes( CaseIfDanglingElseOutterComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          if ( 1'd1 ) begin
            if ( 1'd0 ) begin
              out = in_1;
            end
          end
          else
            out = in_2;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            if ( 1'd1 ) begin
              if ( 1'd0 ) begin
                out = in_1;
              end
            end
            else
              out = in_2;
          end

        endmodule
    '''
)

CaseElifBranchComp = set_attributes( CaseElifBranchComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          if ( 1'd1 ) begin
            out = in_1;
          end
          else if ( 1'd0 ) begin
            out = in_2;
          end
          else
            out = in_3;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          input logic [31:0] in_3,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            if ( 1'd1 ) begin
              out = in_1;
            end
            else if ( 1'd0 ) begin
              out = in_2;
            end
            else
              out = in_3;
          end

        endmodule
    '''
)

CaseNestedIfComp = set_attributes( CaseNestedIfComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          if ( 1'd1 ) begin
            if ( 1'd0 ) begin
              out = in_1;
            end
            else
              out = in_2;
          end
          else if ( 1'd0 ) begin
            if ( 1'd1 ) begin
              out = in_2;
            end
            else
              out = in_3;
          end
          else if ( 1'd1 ) begin
            out = in_3;
          end
          else
            out = in_1;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          input logic [31:0] in_3,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            if ( 1'd1 ) begin
              if ( 1'd0 ) begin
                out = in_1;
              end
              else
                out = in_2;
            end
            else if ( 1'd0 ) begin
              if ( 1'd1 ) begin
                out = in_2;
              end
              else
                out = in_3;
            end
            else if ( 1'd1 ) begin
              out = in_3;
            end
            else
              out = in_1;
          end

        endmodule
    '''
)

CaseForLoopEmptySequenceComp = set_attributes( CaseForLoopEmptySequenceComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = 4'd15;
          for ( int unsigned i = 1'd0; i < 1'd0; i += 1'd1 )
            out[2'(i)] = 1'd0;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [3:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = 4'd15;
            for ( int unsigned i = 1'd0; i < 1'd0; i += 1'd1 )
              out[2'(i)] = 1'd0;
          end

        endmodule
    '''
)

CaseForRangeLowerUpperStepPassThroughComp = set_attributes( CaseForRangeLowerUpperStepPassThroughComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int unsigned i = 1'd0; i < 3'd5; i += 2'd2 )
            out[3'(i)] = in_[3'(i)];
          for ( int unsigned i = 1'd1; i < 3'd5; i += 2'd2 )
            out[3'(i)] = in_[3'(i)];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out [0:4],
          input logic [0:0] reset
        );

          always_comb begin : upblk
            for ( int unsigned i = 1'd0; i < 3'd5; i += 2'd2 )
              out[3'(i)] = in_[3'(i)];
            for ( int unsigned i = 1'd1; i < 3'd5; i += 2'd2 )
              out[3'(i)] = in_[3'(i)];
          end

        endmodule
    '''
)

CaseIfExpBothImplicitComp = set_attributes( CaseIfExpBothImplicitComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_ ? 32'd1 : 32'd0;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = in_ ? 32'd1 : 32'd0;
          end

        endmodule
    '''
)

CaseIfExpInForStmtComp = set_attributes( CaseIfExpInForStmtComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int unsigned i = 1'd0; i < 3'd5; i += 1'd1 )
            out[3'(i)] = ( 3'(i) == 3'd1 ) ? in_[3'(i)] : in_[3'd0];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out [0:4],
          input logic [0:0] reset
        );

          always_comb begin : upblk
            for ( int unsigned i = 1'd0; i < 3'd5; i += 1'd1 )
              out[3'(i)] = ( 3'(i) == 3'd1 ) ? in_[3'(i)] : in_[3'd0];
          end

        endmodule
    '''
)

CaseIfExpUnaryOpInForStmtComp = set_attributes( CaseIfExpUnaryOpInForStmtComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int unsigned i = 0; i < 5; i += 1 )
            out[i] = ( i == 1 ) ? ~in_[i] : in_[0];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out [0:4],
          input logic [0:0] reset
        );

          always_comb begin : upblk
            for ( int unsigned i = 0; i < 5; i += 1 )
              out[i] = ( i == 1 ) ? ~in_[i] : in_[0];
          end

        endmodule
    '''
)

CaseIfBoolOpInForStmtComp = set_attributes( CaseIfBoolOpInForStmtComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int unsigned i = 1'd0; i < 3'd5; i += 1'd1 )
            if ( ( in_[3'(i)] != 32'd0 ) & ( ( 3'(i) < 3'd4 ) ? in_[3'(i) + 3'd1] != 32'd0 : in_[3'd4] != 32'd0 ) ) begin
              out[3'(i)] = in_[3'(i)];
            end
            else
              out[3'(i)] = 32'd0;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out [0:4],
          input logic [0:0] reset
        );

          always_comb begin : upblk
            for ( int unsigned i = 1'd0; i < 3'd5; i += 1'd1 )
              if ( ( in_[3'(i)] != 32'd0 ) & ( ( 3'(i) < 3'd4 ) ? in_[3'(i) + 3'd1] != 32'd0 : in_[3'd4] != 32'd0 ) ) begin
                out[3'(i)] = in_[3'(i)];
              end
              else
                out[3'(i)] = 32'd0;
          end

        endmodule
    '''
)

CaseIfTmpVarInForStmtComp = set_attributes( CaseIfTmpVarInForStmtComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int unsigned i = 1'd0; i < 3'd5; i += 1'd1 ) begin
            if ( ( in_[3'(i)] != 32'd0 ) & ( ( 3'(i) < 3'd4 ) ? in_[3'(i) + 3'd1] != 32'd0 : in_[3'd4] != 32'd0 ) ) begin
              __tmpvar__upblk_tmpvar = in_[3'(i)];
            end
            else
              __tmpvar__upblk_tmpvar = 32'd0;
            out[3'(i)] = __tmpvar__upblk_tmpvar;
          end
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out [0:4],
          input logic [0:0] reset
        );
          logic [31:0] __tmpvar__upblk_tmpvar;

          always_comb begin : upblk
            for ( int unsigned i = 1'd0; i < 3'd5; i += 1'd1 ) begin
              if ( ( in_[3'(i)] != 32'd0 ) & ( ( 3'(i) < 3'd4 ) ? in_[3'(i) + 3'd1] != 32'd0 : in_[3'd4] != 32'd0 ) ) begin
                __tmpvar__upblk_tmpvar = in_[3'(i)];
              end
              else
                __tmpvar__upblk_tmpvar = 32'd0;
              out[3'(i)] = __tmpvar__upblk_tmpvar;
            end
          end

        endmodule
    '''
)

CaseInterfaceArrayNonStaticIndexComp = set_attributes( CaseInterfaceArrayNonStaticIndexComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in___foo[in___foo[1'd0][5'd0]];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset,
          input logic [31:0] in___foo [0:1]
        );

          always_comb begin : upblk
            out = in___foo[in___foo[1'd0][5'd0]];
          end

        endmodule
    '''
)

CaseFixedSizeSliceComp = set_attributes( CaseFixedSizeSliceComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int unsigned i = 1'd0; i < 2'd2; i += 1'd1 )
            out[1'(i)] = in_[4'(i) * 4'd8 +: 8];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [15:0] in_,
          output logic [7:0] out [0:1],
          input logic [0:0] reset
        );

          always_comb begin : upblk
            for ( int unsigned i = 1'd0; i < 2'd2; i += 1'd1 )
              out[1'(i)] = in_[4'(i) * 4'd8 +: 8];
          end

        endmodule
    '''
)

CaseBits32FooInBits32OutComp = set_attributes( CaseBits32FooInBits32OutComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in_.foo;
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo__foo_32;

        module DUT_noparam
        (
          input logic [0:0] clk,
          input Bits32Foo__foo_32 in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = in_.foo;
          end

        endmodule
    '''
)

CaseConstStructInstComp = set_attributes( CaseConstStructInstComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = 32'd0;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = 32'd0;
          end

        endmodule
    '''
)

CaseStructPackedArrayUpblkComp = set_attributes( CaseStructPackedArrayUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_.foo[3'd0], in_.foo[3'd1], in_.foo[3'd2] };
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [4:0][31:0] foo;
        } Bits32x5Foo__foo_32x5;

        module DUT_noparam
        (
          input logic [0:0] clk,
          input Bits32x5Foo__foo_32x5 in_,
          output logic [95:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { in_.foo[3'd0], in_.foo[3'd1], in_.foo[3'd2] };
          end

        endmodule
    '''
)

CaseNestedStructPackedArrayUpblkComp = set_attributes( CaseNestedStructPackedArrayUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_.bar[1'd0], in_.woo.foo, in_.foo };
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo__foo_32;

        typedef struct packed {
          logic [31:0] foo;
          logic [1:0][31:0] bar;
          Bits32Foo__foo_32 woo;
        } NestedStructPackedPlusScalar__c467caf2a4dfbfb2;

        module DUT_noparam
        (
          input logic [0:0] clk,
          input NestedStructPackedPlusScalar__c467caf2a4dfbfb2 in_,
          output logic [95:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { in_.bar[1'd0], in_.woo.foo, in_.foo };
          end

        endmodule
    '''
)

CaseConnectValRdyIfcUpblkComp = set_attributes( CaseConnectValRdyIfcUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out__val = in___val;
          out__msg = in___msg;
          in___rdy = out__rdy;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [0:0] reset,
          input logic [31:0] in___msg,
          output logic [0:0] in___rdy,
          input logic [0:0] in___val,
          output logic [31:0] out__msg,
          input logic [0:0] out__rdy,
          output logic [0:0] out__val
        );

          always_comb begin : upblk
            out__val = in___val;
            out__msg = in___msg;
            in___rdy = out__rdy;
          end

        endmodule
    '''
)

CaseArrayBits32IfcInUpblkComp = set_attributes( CaseArrayBits32IfcInUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in___foo[3'd1];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset,
          input logic [31:0] in___foo [0:4]
        );

          always_comb begin : upblk
            out = in___foo[3'd1];
          end

        endmodule
    '''
)

CaseBits32SubCompAttrUpblkComp = set_attributes( CaseBits32SubCompAttrUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = b__out;
        end
    ''',
    'REF_SRC',
    '''\
        module Bits32OutDrivenComp_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule

        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          logic [0:0] b__clk;
          logic [31:0] b__out;
          logic [0:0] b__reset;

          Bits32OutDrivenComp_noparam b
          (
            .clk( b__clk ),
            .out( b__out ),
            .reset( b__reset )
          );

          always_comb begin : upblk
            out = b__out;
          end

          assign b__clk = clk;
          assign b__reset = reset;

        endmodule
    '''
)

CaseBits32ArraySubCompAttrUpblkComp = set_attributes( CaseBits32ArraySubCompAttrUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = b__out[3'd1];
        end
    ''',
    'REF_SRC',
    '''\
        module Bits32OutDrivenComp_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule

        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          logic [0:0] b__clk [0:4] ;
          logic [31:0] b__out [0:4] ;
          logic [0:0] b__reset [0:4] ;

          Bits32OutDrivenComp_noparam b__0
          (
            .clk( b__clk[0] ),
            .out( b__out[0] ),
            .reset( b__reset[0] )
          );

          Bits32OutDrivenComp_noparam b__1
          (
            .clk( b__clk[1] ),
            .out( b__out[1] ),
            .reset( b__reset[1] )
          );

          Bits32OutDrivenComp_noparam b__2
          (
            .clk( b__clk[2] ),
            .out( b__out[2] ),
            .reset( b__reset[2] )
          );

          Bits32OutDrivenComp_noparam b__3
          (
            .clk( b__clk[3] ),
            .out( b__out[3] ),
            .reset( b__reset[3] )
          );

          Bits32OutDrivenComp_noparam b__4
          (
            .clk( b__clk[4] ),
            .out( b__out[4] ),
            .reset( b__reset[4] )
          );

          always_comb begin : upblk
            out = b__out[3'd1];
          end

          assign b__clk[0] = clk;
          assign b__reset[0] = reset;
          assign b__clk[1] = clk;
          assign b__reset[1] = reset;
          assign b__clk[2] = clk;
          assign b__reset[2] = reset;
          assign b__clk[3] = clk;
          assign b__reset[3] = reset;
          assign b__clk[4] = clk;
          assign b__reset[4] = reset;

        endmodule
    '''
)

CaseConnectInToWireComp = set_attributes( CaseConnectInToWireComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_ [0:4],
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '''\
        logic [31:0] wire_ [0:4];
    ''',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = wire_[2];
        assign wire_[0] = in_[0];
        assign wire_[1] = in_[1];
        assign wire_[2] = in_[2];
        assign wire_[3] = in_[3];
        assign wire_[4] = in_[4];
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_ [0:4],
          output logic [31:0] out,
          input logic [0:0] reset
        );
          logic [31:0] wire_ [0:4];

          assign out = wire_[2];
          assign wire_[0] = in_[0];
          assign wire_[1] = in_[1];
          assign wire_[2] = in_[2];
          assign wire_[3] = in_[3];
          assign wire_[4] = in_[4];

        endmodule
    '''
)

CaseConnectBitsConstToOutComp = set_attributes( CaseConnectBitsConstToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = 32'd0;
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd0;

        endmodule
    '''
)

CaseConnectConstToOutComp = set_attributes( CaseConnectConstToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '''\
        localparam logic [5:0] const_ [0:4] = '{ 6'd42, 6'd42, 6'd42, 6'd42, 6'd42 };
    ''',
    'REF_CONN',
    '''\
        assign out = 32'd42;
    ''',
    'REF_SRC',
    # const_ is not used in upblks so will be optimized away
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule
    '''
)

CaseConnectBitSelToOutComp = set_attributes( CaseConnectBitSelToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_,
        output logic [0:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = in_[0:0];
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [0:0] out,
          input logic [0:0] reset
        );

          assign out = in_[0:0];

        endmodule
    '''
)

CaseConnectSliceToOutComp = set_attributes( CaseConnectSliceToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_,
        output logic [3:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = in_[7:4];
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [3:0] out,
          input logic [0:0] reset
        );

          assign out = in_[7:4];

        endmodule
    '''
)

CaseBitSelOverBitSelComp = set_attributes( CaseBitSelOverBitSelComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_,
        output logic [0:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = in_[1:1];
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [0:0] out,
          input logic [0:0] reset
        );

          assign out = in_[1:1];

        endmodule
    '''
)

CaseBitSelOverPartSelComp = set_attributes( CaseBitSelOverPartSelComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_,
        output logic [0:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = in_[0:0];
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [0:0] out,
          input logic [0:0] reset
        );

          assign out = in_[0:0];

        endmodule
    '''
)

CasePartSelOverBitSelComp = set_attributes( CasePartSelOverBitSelComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_,
        output logic [0:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = in_[1:1];
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [0:0] out,
          input logic [0:0] reset
        );

          assign out = in_[1:1];

        endmodule
    '''
)

CasePartSelOverPartSelComp = set_attributes( CasePartSelOverPartSelComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input logic [31:0] in_,
        output logic [0:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONST',
    '',
    'REF_CONN',
    '''\
        assign out = in_[0:0];
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [0:0] out,
          input logic [0:0] reset
        );

          assign out = in_[0:0];

        endmodule
    '''
)

CaseConnectConstStructAttrToOutComp = set_attributes( CaseConnectConstStructAttrToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONN',
    '''\
        assign out = 32'd42;
    ''',
    'REF_STRUCT',
    (
        rdt.Struct(Bits32Foo, {'foo':rdt.Vector(32)}),
        dedent('''\
                  typedef struct packed {
                    logic [31:0] foo;
                  } Bits32Foo__foo_32;
               ''')
    ),
    'REF_SRC',
    # struct definition will be eliminated because it's not used
    # in an upblk
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule
    '''
)

CaseConnectLiteralStructComp = set_attributes( CaseConnectLiteralStructComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        output NestedStructPackedPlusScalar__c467caf2a4dfbfb2 out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONN',
    '''\
        assign out = { 32'd42, { { 32'd2, 32'd1 } }, { 32'd3 } };
    ''',
    'REF_STRUCT',
    (
        rdt.Struct(NestedStructPackedPlusScalar, {
          'foo':rdt.Vector(32),
          'bar':rdt.PackedArray([2], rdt.Vector(32)),
          'woo':rdt.Struct(Bits32Foo, {'foo':rdt.Vector(32)}),
        }),
        dedent('''\
                  typedef struct packed {
                    logic [31:0] foo;
                    logic [1:0][31:0] bar;
                    Bits32Foo__foo_32 woo;
                  } NestedStructPackedPlusScalar__c467caf2a4dfbfb2;
               ''')
    ),
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo__foo_32;

        typedef struct packed {
          logic [31:0] foo;
          logic [1:0][31:0] bar;
          Bits32Foo__foo_32 woo;
        } NestedStructPackedPlusScalar__c467caf2a4dfbfb2;

        module DUT_noparam
        (
          input logic [0:0] clk,
          output NestedStructPackedPlusScalar__c467caf2a4dfbfb2 out,
          input logic [0:0] reset
        );

          assign out = { 32'd42, { { 32'd2, 32'd1 } }, { 32'd3 } };

        endmodule
    '''
)

CaseConnectArrayStructAttrToOutComp = set_attributes( CaseConnectArrayStructAttrToOutComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input Bits32x5Foo__foo_32x5 in_,
        output logic [31:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONN',
    '''\
        assign out = in_.foo[1];
    ''',
    'REF_STRUCT',
    (
        rdt.Struct(Bits32x5Foo, {'foo':rdt.PackedArray([5], rdt.Vector(32))}),
        dedent('''\
                  typedef struct packed {
                    logic [4:0][31:0] foo;
                  } Bits32x5Foo__foo_32x5;
               ''')
    ),
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [4:0][31:0] foo;
        } Bits32x5Foo__foo_32x5;

        module DUT_noparam
        (
          input logic [0:0] clk,
          input Bits32x5Foo__foo_32x5 in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = in_.foo[1];

        endmodule
    '''
)

CaseConnectNestedStructPackedArrayComp = set_attributes( CaseConnectNestedStructPackedArrayComp,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        input NestedStructPackedPlusScalar__c467caf2a4dfbfb2 in_,
        output logic [95:0] out,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '',
    'REF_CONN',
    '''\
        assign out[31:0] = in_.foo;
        assign out[63:32] = in_.woo.foo;
        assign out[95:64] = in_.bar[0];
    ''',
    'REF_STRUCT',
    (
        rdt.Struct(NestedStructPackedPlusScalar, {
          'foo':rdt.Vector(32),
          'bar':rdt.PackedArray([2], rdt.Vector(32)),
          'woo':rdt.Struct(Bits32Foo, {'foo':rdt.Vector(32)}),
        }),
        dedent('''\
                  typedef struct packed {
                    logic [31:0] foo;
                    logic [1:0][31:0] bar;
                    Bits32Foo__foo_32 woo;
                  } NestedStructPackedPlusScalar__c467caf2a4dfbfb2;
               ''')
    ),
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo__foo_32;

        typedef struct packed {
          logic [31:0] foo;
          logic [1:0][31:0] bar;
          Bits32Foo__foo_32 woo;
        } NestedStructPackedPlusScalar__c467caf2a4dfbfb2;

        module DUT_noparam
        (
          input logic [0:0] clk,
          input NestedStructPackedPlusScalar__c467caf2a4dfbfb2 in_,
          output logic [95:0] out,
          input logic [0:0] reset
        );

          assign out[31:0] = in_.foo;
          assign out[63:32] = in_.woo.foo;
          assign out[95:64] = in_.bar[0];

        endmodule
    '''
)

CaseConnectValRdyIfcComp = set_attributes( CaseConnectValRdyIfcComp,
    'REF_IFC',
    '''\
        input  logic [31:0] in___msg,
        output logic [0:0]  in___rdy,
        input  logic [0:0]  in___val,
        output logic [31:0] out__msg,
        input  logic [0:0]  out__rdy,
        output logic [0:0]  out__val
    ''',
    'REF_CONN',
    '''\
        assign out__msg = in___msg;
        assign in___rdy = out__rdy;
        assign out__val = in___val;
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [0:0] reset,
          input logic [31:0] in___msg,
          output logic [0:0] in___rdy,
          input logic [0:0] in___val,
          output logic [31:0] out__msg,
          input logic [0:0] out__rdy,
          output logic [0:0] out__val
        );

          assign out__msg = in___msg;
          assign in___rdy = out__rdy;
          assign out__val = in___val;

        endmodule
    '''
)

CaseConnectArrayBits32FooIfcComp = set_attributes( CaseConnectArrayBits32FooIfcComp,
    'REF_IFC',
    '''\
        input Bits32Foo__foo_32  in___foo [0:1],
        output Bits32Foo__foo_32  out__foo [0:1]
    ''',
    'REF_CONN',
    '''\
        assign out__foo[0] = in___foo[0];
        assign out__foo[1] = in___foo[1];
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo;
        } Bits32Foo__foo_32;

        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [0:0] reset,
          input Bits32Foo__foo_32 in___foo [0:1],
          output Bits32Foo__foo_32 out__foo [0:1]
        );

          assign out__foo[0] = in___foo[0];
          assign out__foo[1] = in___foo[1];

        endmodule
    '''
)

CaseConnectArrayNestedIfcComp = set_attributes( CaseConnectArrayNestedIfcComp,
    'REF_IFC',
    '''\
        input  logic [0:0]  in___ctrl_foo [0:1],
        input  logic [31:0] in___memifc__msg [0:1],
        output logic [0:0]  in___memifc__rdy [0:1],
        input  logic [0:0]  in___memifc__val [0:1],

        output  logic [0:0]  out__ctrl_foo [0:1],
        output  logic [31:0] out__memifc__msg [0:1],
        input   logic [0:0]  out__memifc__rdy [0:1],
        output  logic [0:0]  out__memifc__val [0:1]
    ''',
    'REF_CONN',
    '''\
        assign out__ctrl_foo[0] = in___ctrl_foo[0];
        assign out__memifc__msg[0] = in___memifc__msg[0];
        assign in___memifc__rdy[0] = out__memifc__rdy[0];
        assign out__memifc__val[0] = in___memifc__val[0];
        assign out__ctrl_foo[1] = in___ctrl_foo[1];
        assign out__memifc__msg[1] = in___memifc__msg[1];
        assign in___memifc__rdy[1] = out__memifc__rdy[1];
        assign out__memifc__val[1] = in___memifc__val[1];
    ''',
    'REF_SRC',
    '''\
        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [0:0] reset,
          input logic [0:0] in___ctrl_foo [0:1],
          input logic [31:0] in___memifc__msg [0:1],
          output logic [0:0] in___memifc__rdy [0:1],
          input logic [0:0] in___memifc__val [0:1],
          output logic [0:0] out__ctrl_foo [0:1],
          output logic [31:0] out__memifc__msg [0:1],
          input logic [0:0] out__memifc__rdy [0:1],
          output logic [0:0] out__memifc__val [0:1]
        );

          assign out__ctrl_foo[0] = in___ctrl_foo[0];
          assign out__memifc__msg[0] = in___memifc__msg[0];
          assign in___memifc__rdy[0] = out__memifc__rdy[0];
          assign out__memifc__val[0] = in___memifc__val[0];
          assign out__ctrl_foo[1] = in___ctrl_foo[1];
          assign out__memifc__msg[1] = in___memifc__msg[1];
          assign in___memifc__rdy[1] = out__memifc__rdy[1];
          assign out__memifc__val[1] = in___memifc__val[1];

        endmodule
    '''
)

CaseBits32ConnectSubCompAttrComp = set_attributes( CaseBits32ConnectSubCompAttrComp,
    'REF_COMP',
    '''\
        logic [0:0] b__clk;
        logic [31:0] b__out;
        logic [0:0] b__reset;

        Bits32OutDrivenComp_noparam b
        (
          .clk( b__clk ),
          .out( b__out ),
          .reset( b__reset )
        );
    ''',
    'REF_SRC',
    '''\
        module Bits32OutDrivenComp_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule

        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          logic [0:0] b__clk;
          logic [31:0] b__out;
          logic [0:0] b__reset;

          Bits32OutDrivenComp_noparam b
          (
            .clk( b__clk ),
            .out( b__out ),
            .reset( b__reset )
          );

          assign b__clk = clk;
          assign b__reset = reset;
          assign out = b__out;

        endmodule
    '''
)

CaseConnectArraySubCompArrayStructIfcComp = set_attributes( CaseConnectArraySubCompArrayStructIfcComp,
    'REF_COMP',
    '''\
      logic [0:0] b__clk [0:0] ;
      logic [31:0] b__out [0:0] ;
      logic [0:0] b__reset [0:0] ;
      Bits32Foo__foo_32 b__ifc__foo [0:0][0:0][0:0] ;

      Bits32ArrayStructIfcComp_noparam b__0
      (
        .clk( b__clk[0] ),
        .out( b__out[0] ),
        .reset( b__reset[0] ),
        .ifc__foo( b__ifc__foo[0] )
      );
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo ;
        } Bits32Foo__foo_32;

        module Bits32ArrayStructIfcComp_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset,
          input Bits32Foo__foo_32 ifc__foo [0:0][0:0]
        );

          assign out = ifc__foo[0][0].foo;

        endmodule

        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          logic [0:0] b__clk [0:0] ;
          logic [31:0] b__out [0:0] ;
          logic [0:0] b__reset [0:0] ;
          Bits32Foo__foo_32 b__ifc__foo [0:0][0:0][0:0] ;

          Bits32ArrayStructIfcComp_noparam b__0
          (
            .clk( b__clk[0] ),
            .out( b__out[0] ),
            .reset( b__reset[0] ),
            .ifc__foo( b__ifc__foo[0] )
          );

          assign b__clk[0] = clk;
          assign b__reset[0] = reset;
          assign b__ifc__foo[0][0][0].foo = in_;
          assign out = b__out[0];

        endmodule
    '''
)

CaseBits32ArrayConnectSubCompAttrComp = set_attributes( CaseBits32ArrayConnectSubCompAttrComp,
    'REF_COMP',
    '''\
      logic [0:0] b__clk [0:4] ;
      logic [31:0] b__out [0:4] ;
      logic [0:0] b__reset [0:4] ;

      Bits32OutDrivenComp_noparam b__0
      (
        .clk( b__clk[0] ),
        .out( b__out[0] ),
        .reset( b__reset[0] )
      );

      Bits32OutDrivenComp_noparam b__1
      (
        .clk( b__clk[1] ),
        .out( b__out[1] ),
        .reset( b__reset[1] )
      );

      Bits32OutDrivenComp_noparam b__2
      (
        .clk( b__clk[2] ),
        .out( b__out[2] ),
        .reset( b__reset[2] )
      );

      Bits32OutDrivenComp_noparam b__3
      (
        .clk( b__clk[3] ),
        .out( b__out[3] ),
        .reset( b__reset[3] )
      );

      Bits32OutDrivenComp_noparam b__4
      (
        .clk( b__clk[4] ),
        .out( b__out[4] ),
        .reset( b__reset[4] )
      );
    ''',
    'REF_SRC',
    '''\
        module Bits32OutDrivenComp_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule

        module DUT_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          logic [0:0] b__clk [0:4] ;
          logic [31:0] b__out [0:4] ;
          logic [0:0] b__reset [0:4] ;

          Bits32OutDrivenComp_noparam b__0
          (
            .clk( b__clk[0] ),
            .out( b__out[0] ),
            .reset( b__reset[0] )
          );

          Bits32OutDrivenComp_noparam b__1
          (
            .clk( b__clk[1] ),
            .out( b__out[1] ),
            .reset( b__reset[1] )
          );

          Bits32OutDrivenComp_noparam b__2
          (
            .clk( b__clk[2] ),
            .out( b__out[2] ),
            .reset( b__reset[2] )
          );

          Bits32OutDrivenComp_noparam b__3
          (
            .clk( b__clk[3] ),
            .out( b__out[3] ),
            .reset( b__reset[3] )
          );

          Bits32OutDrivenComp_noparam b__4
          (
            .clk( b__clk[4] ),
            .out( b__out[4] ),
            .reset( b__reset[4] )
          );

          assign b__clk[0] = clk;
          assign b__reset[0] = reset;
          assign b__clk[1] = clk;
          assign b__reset[1] = reset;
          assign b__clk[2] = clk;
          assign b__reset[2] = reset;
          assign b__clk[3] = clk;
          assign b__reset[3] = reset;
          assign b__clk[4] = clk;
          assign b__reset[4] = reset;
          assign out = b__out[1];

        endmodule
    '''
)

CasePlaceholderTranslationVReg = set_attributes( CasePlaceholderTranslationVReg,
    'REF_PORT',
    '''\
        input logic [0:0] clk,
        output logic [31:0] d,
        input logic [31:0] q,
        input logic [0:0] reset
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONST',
    '''\
    ''',
    'REF_CONN',
    '''\
    ''',
    'REF_PLACEHOLDER_WRAPPER',
    '''\
        // This is a top-level module that wraps a parametrized module
        // This file is generated by PyMTL SystemVerilog import pass
        module placeholder_VReg
        (
          input logic [1-1:0] clk,
          output logic [32-1:0] d,
          input logic [32-1:0] q,
          input logic [1-1:0] reset
        );
          VReg
          #(
          ) v
          (
            .clk( clk ),
            .d( d ),
            .q( q ),
            .reset( reset )
          );
        endmodule
    ''',
    'REF_PLACEHOLDER_DEPENDENCY',
    '''\
        `line 1 "VReg.v" 0
        module VReg(
          input  logic          clk,
          input  logic          reset,
          output logic [32-1:0] q,
          input  logic [32-1:0] d
        );
          always_ff @(posedge clk) begin
            q <= d;
          end

        endmodule
    ''',
    'REF_SRC',
    '''\
        // PyMTL Placeholder Bits32VRegComp Definition

        // Here are the modules that Bits32VRegComp depends on
        `line 1 "VReg.v" 0
        module VReg(
          input  logic          clk,
          input  logic          reset,
          output logic [32-1:0] q,
          input  logic [32-1:0] d
        );
          always_ff @(posedge clk) begin
            q <= d;
          end

        endmodule

        // Here is the wrapper of the Verilog module backing the
        // placeholder
        module placeholder_VReg
        (
          input logic [1-1:0] clk,
          output logic [32-1:0] d,
          input logic [32-1:0] q,
          input logic [1-1:0] reset
        );
          VReg
          #(
          ) v
          (
            .clk( clk ),
            .d( d ),
            .q( q ),
            .reset( reset )
          );
        endmodule
    ''',
)

CasePlaceholderTranslationRegIncr = set_attributes( CasePlaceholderTranslationRegIncr,
    'REF_PORT',
    '''\
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONST',
    '''\
    ''',
    'REF_CONN',
    '''\
    ''',
    'REF_PLACEHOLDER_WRAPPER',
    '''\
    ''',
    'REF_PLACEHOLDER_DEPENDENCY',
    '''\
    ''',
)

CaseHeteroCompArrayComp = set_attributes( CaseHeteroCompArrayComp,
    'REF_COMP',
    '''\
        logic [0:0] comps__clk [0:1] ;
        logic [31:0] comps__in_ [0:1] ;
        logic [31:0] comps__out [0:1] ;
        logic [0:0] comps__reset [0:1] ;

        Bits32DummyFooComp_noparam comps__0
        (
          .clk( comps__clk[0] ),
          .in_( comps__in_[0] ),
          .out( comps__out[0] ),
          .reset( comps__reset[0] )
        );

        Bits32DummyBarComp_noparam comps__1
        (
          .clk( comps__clk[1] ),
          .in_( comps__in_[1] ),
          .out( comps__out[1] ),
          .reset( comps__reset[1] )
        );
    ''',
    'REF_SRC',
    '''\
        module Bits32DummyFooComp_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          assign out = in_;
        endmodule

        module Bits32DummyBarComp_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          always_comb begin : upblk
            out = in_ + 32'd42;
          end
        endmodule

        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_1,
          input logic [31:0] in_2,
          output logic [31:0] out_1,
          output logic [31:0] out_2,
          input logic [0:0] reset
        );
          logic [0:0] comps__clk [0:1] ;
          logic [31:0] comps__in_ [0:1] ;
          logic [31:0] comps__out [0:1] ;
          logic [0:0] comps__reset [0:1] ;

          Bits32DummyFooComp_noparam comps__0
          (
            .clk( comps__clk[0] ),
            .in_( comps__in_[0] ),
            .out( comps__out[0] ),
            .reset( comps__reset[0] )
          );

          Bits32DummyBarComp_noparam comps__1
          (
            .clk( comps__clk[1] ),
            .in_( comps__in_[1] ),
            .out( comps__out[1] ),
            .reset( comps__reset[1] )
          );

          assign comps__clk[0] = clk;
          assign comps__reset[0] = reset;
          assign comps__clk[1] = clk;
          assign comps__reset[1] = reset;
          assign comps__in_[0] = in_1;
          assign comps__in_[1] = in_2;
          assign out_1 = comps__out[0];
          assign out_2 = comps__out[1];

        endmodule
    '''
)

CaseChildExplicitModuleName = set_attributes( CaseChildExplicitModuleName,
    'REF_COMP',
    '''\
        logic [0:0] child__clk ;
        logic [31:0] child__in_ ;
        logic [0:0] child__reset ;

        NewChildDUTName child
        (
          .clk( child__clk ),
          .in_( child__in_ ),
          .reset( child__reset )
        );
    ''',
)

CaseBehavioralArraySubCompArrayStructIfcComp = set_attributes( CaseBehavioralArraySubCompArrayStructIfcComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          for ( int unsigned i = 1'd0; i < 2'd2; i += 1'd1 )
            for ( int unsigned j = 1'd0; j < 1'd1; j += 1'd1 )
              b__ifc__foo[1'(i)][1'(j)][1'd0].foo = in_;
          out = b__out[1'd1];
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo ;
        } Bits32Foo__foo_32;

        module Bits32ArrayStructIfcComp_noparam
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset,
          input Bits32Foo__foo_32 ifc__foo [0:0][0:0]
        );

          assign out = ifc__foo[0][0].foo;

        endmodule

        module DUT_noparam
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          logic [0:0] b__clk [0:1] ;
          logic [31:0] b__out [0:1] ;
          logic [0:0] b__reset [0:1] ;
          Bits32Foo__foo_32 b__ifc__foo [0:1][0:0][0:0] ;

          Bits32ArrayStructIfcComp_noparam b__0
          (
            .clk( b__clk[0] ),
            .out( b__out[0] ),
            .reset( b__reset[0] ),
            .ifc__foo( b__ifc__foo[0] )
          );

          Bits32ArrayStructIfcComp_noparam b__1
          (
            .clk( b__clk[1] ),
            .out( b__out[1] ),
            .reset( b__reset[1] ),
            .ifc__foo( b__ifc__foo[1] )
          );

          always_comb begin : upblk
            for ( int unsigned i = 1'd0; i < 2'd2; i += 1'd1 )
              for ( int unsigned j = 1'd0; j < 1'd1; j += 1'd1 )
                b__ifc__foo[1'(i)][1'(j)][1'd0].foo = in_;
            out = b__out[1'd1];
          end

          assign b__clk[0] = clk;
          assign b__reset[0] = reset;
          assign b__clk[1] = clk;
          assign b__reset[1] = reset;

        endmodule
    '''
)

CaseBits32FooNoArgBehavioralComp = set_attributes( CaseBits32FooNoArgBehavioralComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = 32'd0;
        end
    ''',
    'REF_SRC',
    '''\
        typedef struct packed {
          logic [31:0] foo ;
        } Bits32Foo__foo_32;

        module DUT_noparam
        (
          input logic [0:0] clk,
          output Bits32Foo__foo_32 out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = 32'd0;
          end

        endmodule
    '''
)



#=========================================================================
# ImportedObject_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 2, 2019
"""Test if the imported object works correctly."""

import gc
from os.path import dirname

import pytest

from pymtl3 import DefaultPassGroup, Interface
from pymtl3.datatypes import Bits1, Bits10, Bits32, Bits48, Bits64, clog2, mk_bits
from pymtl3.dsl import Component, InPort, Interface, OutPort, connect
from pymtl3.passes.backends.verilog import (
    VerilogPlaceholder,
    VerilogPlaceholderPass,
    VerilogTranslationImportPass,
    VerilogVerilatorImportPass,
)
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.test_utils import TestVectorSimulator


def local_do_test( _m ):
  _m.elaborate()
  if not hasattr( _m, "_no_trans_import" ):
    _m.set_metadata( VerilogTranslationImportPass.enable, True )
  _m.apply( VerilogPlaceholderPass() )
  m = VerilogTranslationImportPass()( _m )
  sim = TestVectorSimulator( m, _m._tvs, _m._tv_in, _m._tv_out )
  sim.run_test()

def test_reg( do_test ):
  # General trans-import test
  def tv_in( m, tv ):
    m.in_ @= Bits32( tv[0] )
  def tv_out( m, tv ):
    if tv[1] != '*':
      assert m.out == Bits32( tv[1] )
  class VReg( Component, VerilogPlaceholder ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.set_metadata( VerilogPlaceholderPass.port_map, {
          s.clk : "clk", s.reset : "reset",
          s.in_ : "d",   s.out : "q",
      } )
  a = VReg()
  a._tvs = [
    [    1,    '*' ],
    [    2,      1 ],
    [   -1,      2 ],
    [   -2,     -1 ],
    [   42,     -2 ],
    [  -42,     42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_vl_uninit( do_test ):
  # Use a latch to test if verilator has correctly set up
  # the inital signal values
  def tv_in( m, tv ):
    m.in_ @= Bits32( tv[0] )
  def tv_out( m, tv ):
    assert m.out == Bits32( tv[1] )
  class VUninit( Component, VerilogPlaceholder ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.set_metadata( VerilogPlaceholderPass.port_map, {
          s.in_ : "d", s.out : "q",
      } )
      s.set_metadata( VerilogVerilatorImportPass.vl_xinit, 'ones' )
  a = VUninit()
  a._tvs = [
    [    0, 4294967295 ],
    [    2, 4294967295 ],
    [   42,         42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_reg_incomplete_portmap( do_test ):
  # Test support for incomplete port map
  def tv_in( m, tv ):
    m.in_ @= Bits32( tv[0] )
  def tv_out( m, tv ):
    if tv[1] != '*':
      assert m.out == Bits32( tv[1] )
  class VReg( Component, VerilogPlaceholder ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )

      s.set_metadata( VerilogPlaceholderPass.port_map, {
          s.in_ : "d",   s.out : "q",
      } )
  a = VReg()
  a._tvs = [
    [    1,    '*' ],
    [    2,      1 ],
    [   -1,      2 ],
    [   -2,     -1 ],
    [   42,     -2 ],
    [  -42,     42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_adder( do_test ):
  # Test an adder
  def tv_in( m, tv ):
    m.in0 @= Bits32( tv[0] )
    m.in1 @= Bits32( tv[1] )
    m.cin @= Bits1(  tv[2] )
  def tv_out( m, tv ):
    if tv[3] != '*':
      assert m.out == Bits32( tv[3] )
    if tv[4] != '*':
      assert m.cout == Bits1( tv[4] )
  class VAdder( Component, VerilogPlaceholder ):
    def construct( s ):
      s.in0 = InPort( Bits32 )
      s.in1 = InPort( Bits32 )
      s.cin = InPort( Bits1 )
      s.out = OutPort( Bits32 )
      s.cout = OutPort( Bits1 )
  a = VAdder()
  a._tvs = [
    [    1,      1,     1,     3, 0 ],
    [    1,     -1,     0,     0, 1 ],
    [   42,     42,     1,    85, 0 ],
    [   42,    -43,     1,     0, 1 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_normal_queue_implicit_top_module( do_test ):
  # Test a Placeholder with params in `construct`
  def tv_in( m, tv ):
    m.enq_en @= Bits1( tv[0] )
    m.enq_msg @= Bits32( tv[1] )
    m.deq_en @= Bits1( tv[3] )
  def tv_out( m, tv ):
    if tv[2] != '*':
      assert m.enq_rdy == Bits1( tv[2] )
    if tv[4] != '*':
      assert m.deq_rdy == Bits1( tv[5] )
    if tv[5] != '*':
      assert m.deq_msg == Bits32( tv[4] )
  class VQueue( Component, VerilogPlaceholder ):
    def construct( s, data_width, num_entries, count_width ):
      s.count   =  OutPort( mk_bits( count_width )  )
      s.deq_en  =  InPort( Bits1  )
      s.deq_rdy = OutPort( Bits1  )
      s.deq_msg = OutPort( mk_bits( data_width ) )
      s.enq_en  =  InPort( Bits1  )
      s.enq_rdy = OutPort( Bits1  )
      s.enq_msg =  InPort( mk_bits( data_width ) )
  num_entries = 1
  q = VQueue(
      data_width = 32,
      num_entries = num_entries,
      count_width = clog2(num_entries+1))
  tv = [
    #   enq                deq
    #   en    msg   rdy    en    msg   rdy
    [    1,    42,    1,    0,     0,    0  ],
    [    0,    43,    0,    1,    42,    1  ],
    [    1,    43,    1,    0,    42,    0  ],
    [    0,    44,    0,    1,    43,    1  ],
    [    1,    44,    1,    0,    43,    0  ],
    [    0,    45,    0,    1,    44,    1  ],
    [    1,    45,    1,    0,    44,    0  ],
  ]
  q._tvs = tv
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )

def test_normal_queue_params( do_test ):
  # Test the `params` option of placeholder configs
  def tv_in( m, tv ):
    m.enq_en @= Bits1( tv[0] )
    m.enq_msg @= Bits32( tv[1] )
    m.deq_en @= Bits1( tv[3] )
  def tv_out( m, tv ):
    if tv[2] != '*':
      assert m.enq_rdy == Bits1( tv[2] )
    if tv[4] != '*':
      assert m.deq_rdy == Bits1( tv[5] )
    if tv[5] != '*':
      assert m.deq_msg == Bits32( tv[4] )
  class Queue( Component, VerilogPlaceholder ):
    def construct( s, nbits, nelems, nbits_cnt ):
      s.count   =  OutPort( mk_bits( nbits_cnt )  )
      s.deq_en  =  InPort( Bits1  )
      s.deq_rdy = OutPort( Bits1  )
      s.deq_msg = OutPort( mk_bits( nbits ) )
      s.enq_en  =  InPort( Bits1  )
      s.enq_rdy = OutPort( Bits1  )
      s.enq_msg =  InPort( mk_bits( nbits ) )
      s.set_metadata( VerilogPlaceholderPass.src_file, dirname(__file__)+'/VQueue.v' )
      s.set_metadata( VerilogPlaceholderPass.top_module, 'VQueue' )
      s.set_metadata( VerilogPlaceholderPass.params, {
          'data_width'  : nbits,
          'num_entries' : nelems,
          'count_width' : nbits_cnt,
      } )
  num_entries = 1
  q = Queue(
      nbits = 32,
      nelems = num_entries,
      nbits_cnt = clog2(num_entries+1))
  tv = [
    #   enq                deq
    #   en    msg   rdy    en    msg   rdy
    [    1,    42,    1,    0,     0,    0  ],
    [    0,    43,    0,    1,    42,    1  ],
    [    1,    43,    1,    0,    42,    0  ],
    [    0,    44,    0,    1,    43,    1  ],
    [    1,    44,    1,    0,    43,    0  ],
    [    0,    45,    0,    1,    44,    1  ],
    [    1,    45,    1,    0,    44,    0  ],
  ]
  q._tvs = tv
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )

def test_normal_queue_interface( do_test ):
  # Test a Placeholder with params in `construct`
  def tv_in( m, tv ):
    m.enq.en @= Bits1( tv[0] )
    m.enq.msg @= Bits32( tv[1] )
    m.deq.en @= Bits1( tv[3] )
  def tv_out( m, tv ):
    if tv[2] != '*':
      assert m.enq.rdy == Bits1( tv[2] )
    if tv[4] != '*':
      assert m.deq.rdy == Bits1( tv[5] )
    if tv[5] != '*':
      assert m.deq.msg == Bits32( tv[4] )
  class DequeueIfc( Interface ):
    def construct( s, Type ):
      s.en  = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
      s.msg = OutPort( Type )
  class EnqueueIfc( Interface ):
    def construct( s, Type ):
      s.en  = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
      s.msg = InPort( Type )
  class VQueue( Component, VerilogPlaceholder ):
    def construct( s, data_width, num_entries, count_width ):
      s.count = OutPort( mk_bits( count_width )  )
      s.deq   = DequeueIfc( mk_bits( data_width ) )
      s.enq   = EnqueueIfc( mk_bits( data_width ) )
  num_entries = 1
  q = VQueue(
      data_width = 32,
      num_entries = num_entries,
      count_width = clog2(num_entries+1))
  tv = [
    #   enq                deq
    #   en    msg   rdy    en    msg   rdy
    [    1,    42,    1,    0,     0,    0  ],
    [    0,    43,    0,    1,    42,    1  ],
    [    1,    43,    1,    0,    42,    0  ],
    [    0,    44,    0,    1,    43,    1  ],
    [    1,    44,    1,    0,    43,    0  ],
    [    0,    45,    0,    1,    44,    1  ],
    [    1,    45,    1,    0,    44,    0  ],
  ]
  q._tvs = tv
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )

def test_unpacked_port_array( do_test ):
  # Test the `params` option of placeholder configs
  def tv_in( m, tv ):
    m.in_[0] @= Bits32(tv[0])
    m.in_[1] @= Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[2])
    assert m.out[1] == Bits32(tv[3])
  class VPassThrough( Component, VerilogPlaceholder ):
    def construct( s, nports, nbits ):
      s.in_ = [ InPort( mk_bits(nbits) ) for _ in range(nports) ]
      s.out = [ OutPort( mk_bits(nbits) ) for _ in range(nports) ]
      s.set_metadata( VerilogPlaceholderPass.params, {
          'num_ports' : nports,
          'bitwidth'  : nbits,
      } )
      s.set_metadata( VerilogPlaceholderPass.has_clk, False )
      s.set_metadata( VerilogPlaceholderPass.has_reset, False )

  q = VPassThrough( 2, 32 )
  tv = [
    [ 1, 42, 1, 42 ],
    [ 1, -1, 1, -1 ],
    [ 0,  1, 0,  1 ],
    [ -1, 1, -1, 1 ],
  ]
  q._tvs = tv
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )

def test_unpacked_port_array_infer_clk_reset( do_test ):
  # Test the `params` option of placeholder configs
  def tv_in( m, tv ):
    m.in_[0] @= Bits32(tv[0])
    m.in_[1] @= Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[2])
    assert m.out[1] == Bits32(tv[3])
  class VPassThrough( Component, VerilogPlaceholder ):
    def construct( s, nports, nbits ):
      s.in_ = [ InPort( mk_bits(nbits) ) for _ in range(nports) ]
      s.out = [ OutPort( mk_bits(nbits) ) for _ in range(nports) ]
      s.set_metadata( VerilogPlaceholderPass.params, {
          'num_ports' : nports,
          'bitwidth'  : nbits,
      } )

  q = VPassThrough( 2, 32 )
  tv = [
    [ 1, 42, 1, 42 ],
    [ 1, -1, 1, -1 ],
    [ 0,  1, 0,  1 ],
    [ -1, 1, -1, 1 ],
  ]
  q._tvs = tv
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )

@pytest.mark.parametrize(
  "translate", [ True, False ]
)
def test_param_pass_through( do_test, translate ):
  class VPassThrough( Component, VerilogPlaceholder ):
    def construct( s, nports, nbits ):
      s.in_ = [ InPort( mk_bits(nbits) ) for _ in range(nports) ]
      s.out = [ OutPort( mk_bits(nbits) ) for _ in range(nports) ]
      s.set_metadata( VerilogPlaceholderPass.params, {
          'num_ports' : nports,
          'bitwidth'  : nbits,
      } )
      s.set_metadata( VerilogPlaceholderPass.has_clk, False )
      s.set_metadata( VerilogPlaceholderPass.has_reset, False )
  class PassThrough( Component ):
    def construct( s ):
      s.in_ = InPort( Bits48 )
      s.out = OutPort( Bits48 )
      s.pt16 = VPassThrough(1, 16)
      s.pt32 = VPassThrough(1, 32)
      s.pt16.in_[0] //= s.in_[0:16]
      s.pt16.out[0] //= s.out[0:16]
      s.pt32.in_[0] //= s.in_[16:48]
      s.pt32.out[0] //= s.out[16:48]
    def line_trace( s ):
      return f"{s.in_} > {s.out}"
  def tv_in( m, tv ):
    m.in_ @= Bits48( tv[0] )
  def tv_out( m, tv ):
    assert m.out == Bits48( tv[1] )

  p = PassThrough()
  tv = [
    [  1,  1, ],
    [ -1, -1, ],
    [ 42, 42, ],
    [ -2, -2, ],
  ]
  p._tvs = tv
  p._tv_in = tv_in
  p._tv_out = tv_out
  if not translate:
    p._no_trans_import = True
  do_test( p )

def test_non_top_portmap( do_test ):
  def tv_in( m, tv ):
    m.in_ @= Bits32(tv[0])
  def tv_out( m, tv ):
    if tv[1] != '*':
      assert m.out == Bits32(tv[1])
  class VReg( Component, VerilogPlaceholder ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.set_metadata( VerilogPlaceholderPass.port_map, {
          s.in_ : "d", s.out : "q",
      } )
  class Top( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.v = VReg()
      s.v.in_ //= s.in_
      s.v.out //= s.out
  a = Top()
  a._tvs = [
    [    1,    '*' ],
    [    2,      1 ],
    [   -1,      2 ],
    [   -2,     -1 ],
    [   42,     -2 ],
    [  -42,     42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

#-------------------------------------------------------------------------
# test cases that do not use do_test
#-------------------------------------------------------------------------

def test_reg_external_trace( do_test ):
  # Test Verilog line trace
  class VRegTrace( Component, VerilogPlaceholder ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.set_metadata( VerilogPlaceholderPass.port_map, {
          s.in_ : "d", s.out : "q",
      } )
      s.set_metadata( VerilogVerilatorImportPass.vl_line_trace, True )
      s.set_metadata( VerilogTranslationImportPass.enable, True )
  a = VRegTrace()
  a.elaborate()
  a.apply( VerilogPlaceholderPass() )
  a = VerilogTranslationImportPass()( a )
  a.apply( DefaultPassGroup() )

  assert a.line_trace() == 'q =          0'
  a.in_ @= Bits32(1)
  a.sim_tick()
  assert a.line_trace() == 'q =          1'
  a.in_ @= Bits32(2)
  a.sim_tick()
  assert a.line_trace() == 'q =          2'
  a.in_ @= Bits32(-1)
  a.sim_tick()
  # 0xFFFFFFFF unsigned
  assert a.line_trace() == 'q = 4294967295'
  a.sim_tick()
  a.finalize()

def test_reg_infer_external_trace( do_test ):
  # Test Verilog line trace
  class VRegTrace( Component, VerilogPlaceholder ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.set_metadata( VerilogPlaceholderPass.port_map, {
          s.in_ : "d", s.out : "q",
      } )
      s.set_metadata( VerilogTranslationImportPass.enable, True )
  a = VRegTrace()
  a.elaborate()
  a.apply( VerilogPlaceholderPass() )
  a = VerilogTranslationImportPass()( a )
  a.apply( DefaultPassGroup() )

  assert a.line_trace() == 'q =          0'
  a.in_ @= Bits32(1)
  a.sim_tick()
  assert a.line_trace() == 'q =          1'
  a.in_ @= Bits32(2)
  a.sim_tick()
  assert a.line_trace() == 'q =          2'
  a.in_ @= Bits32(-1)
  a.sim_tick()
  # 0xFFFFFFFF unsigned
  assert a.line_trace() == 'q = 4294967295'
  a.sim_tick()
  a.finalize()

def test_incr_on_demand_vcd( do_test ):
  class VIncr( Component, VerilogPlaceholder ):
    def construct( s ):
      s.in_ = InPort( 10 )
      s.out = OutPort( 10 )
      s.vcd_en = OutPort()
      s.set_metadata( VerilogPlaceholderPass.has_clk, False )
      s.set_metadata( VerilogPlaceholderPass.has_reset, False )
      s.set_metadata( VerilogVerilatorImportPass.vl_trace, True )
      s.set_metadata( VerilogVerilatorImportPass.vl_trace_on_demand, True )
      s.set_metadata( VerilogVerilatorImportPass.vl_trace_on_demand_portname, "vcd_en" )
  a = VIncr()
  a.elaborate()
  a.apply( VerilogPlaceholderPass() )
  a = VerilogTranslationImportPass()( a )
  a.apply( DefaultPassGroup() )

  for i in range(64):
    a.in_ @= Bits10(i)
    a.sim_tick()

  a.finalize()



#=========================================================================
# PrintWavePass_test.py
#=========================================================================
# A simple demo of how PrintWavePass is used in the simulation of
# a toy component.
#
# Author : Kaishuo Cheng
# Date   : Oct 8th, 2019

import io
from contextlib import redirect_stdout

from pymtl3.datatypes import (
    Bits1,
    Bits16,
    Bits32,
    Bits128,
    b1,
    b16,
    b32,
    b128,
    bitstruct,
)
from pymtl3.dsl import *
from pymtl3.passes.errors import ModelTypeError
from pymtl3.passes.PassGroups import DefaultPassGroup

from ..PrintTextWavePass import PrintTextWavePass


def test_collect_signal():

  class Toy( Component ):
    """Toy adder component"""

    def construct( s ):
      # Interfaces
      s.in0 = InPort( Bits32 )
      s.in1 = InPort( Bits32 )
      s.out = OutPort( Bits32 )

      @update
      def add_upblk():
        # This update block models the behavior of a 32-bit adder
        s.out @= s.in0 + s.in1

  def process_binary(sig):
    """
    Returns int value from a signal in 32b form. Used for testing.
    Example: input: 0b00000000000000000000000000000000
             output: 0
    """
    tempint = int(sig[2:],2)
    if sig[2] == '1':
       #taking 2's complement.
       #leading 1 indicates a negative number
      return tempint -2 **32

    else:
      return tempint

  # Create a toy component and elaborate it
  dut = Toy()

  dut.elaborate()

  # Turn on textwave
  dut.set_metadata( PrintTextWavePass.enable, True )

  # Setup the simulation
  dut.apply( DefaultPassGroup() )
  dut.sim_reset()

  # Test vector
  vector = [
    #  in0        in1       out
    b32(1),    b32(2),    b32(3),
    b32(-1),   b32(2),    b32(1),
    b32(42),   b32(2),    b32(44),
    b32(1),    b32(-2),   b32(-1),
    b32(42),   b32(-42),  b32(0),
  ]

  testlist = [
  #in0
  [0,0,0,1,-1,42,1,42], #dut's reset is on for the first two cycles".

  #in1
  [0,0,0,2,2,2,-2,-42],

  #out
  [0,0,0,3,1,44,-1,0]
  ]
  # Begin simulation
  for in0, in1, out in zip(vector[0::3], vector[1::3], vector[2::3]):
    dut.in0 @= in0
    dut.in1 @= in1
    dut.sim_tick()
    assert dut.out == out

  #test
  sig = dut.get_metadata( PrintTextWavePass.textwave_dict )
  siglist = ["s.in0","s.in1","s.out","s.reset"]
  for i in siglist:
    assert i in sig,"signals not captured"

  print(sig)
  for i in range(3): # in0, in1, and out
    partsig = sig[ siglist[i] ]
    signal_length = len(partsig)
    assert signal_length >= 5,"missing some cycles of signals"

    for j in range(signal_length):
      assert testlist[i][j] == process_binary(partsig[j]),"collected wrong signals"

  print("All signals captured in top._tracing.text_sigs!")


def test_toy():
  class Toy( Component ):
    """Toy adder component"""

    def construct( s ):
      # Interfaces
      s.i = InPort( Bits32 )
      s.inlong = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.state = Wire(Bits1)
      @update
      def add_upblk():
        # This update block models the behavior of a 32-bit adder
        s.out @= s.i + s.inlong
        if s.out[3]:
          s.state @= s.state +b1(1)
        else:
          s.state @= b1(0)

  # Create a toy component and elaborate it
  dut = Toy()
  dut.set_metadata( PrintTextWavePass.enable, True )

  # Setup the simulation
  dut.apply( DefaultPassGroup() )
  dut.sim_reset()
  # Test vector
  vector = [
    #  i        inlong       out
    b32(1),    b32(2),    b32(3),
    b32(0),    b32(2),    b32(2),
    b32(0),    b32(2),    b32(2),
    b32(1),    b32(-2),   b32(-1),
    b32(1),    b32(-42),  b32(-41),
    b32(1),    b32(-4),   b32(-3),
    b32(1),    b32(2),    b32(3),
    b32(0),    b32(2),    b32(2),
    b32(1),    b32(2),    b32(3),
    b32(0),    b32(-5),   b32(-5),
  ]

  # Begin simulation
  for i, inlong, out in zip(vector[0::3], vector[1::3], vector[2::3]):
    dut.i @= i
    dut.inlong @= inlong
    dut.sim_tick()
    assert dut.out == out

  #print
  f = io.StringIO()
  dut.print_textwave()
  with redirect_stdout(f):
    dut.print_textwave()
  out = f.getvalue()
  for i in dut.get_metadata( PrintTextWavePass.textwave_dict ):
    dot = i.find(".")
    sliced = i[dot+1:]
    if sliced != "reset" and sliced != "clk":
      assert i[dot+1:] in out

def test_widetoy():
  class Toy( Component ):

    def construct( s ):
      # Interfaces
      s.i = InPort( Bits128 )
      s.inlong = InPort( Bits128 )
      s.out = OutPort( Bits128 )
      s.state = Wire(Bits1)
      @update
      def add_upblk():
        # This update block models the behavior of a 32-bit adder
        s.out @= s.i + s.inlong
        if s.out[3]:
          s.state @= s.state + b1(1)
        else:
          s.state @= b1(0)

  # Create a toy component and elaborate it
  dut = Toy()

  dut.set_metadata( PrintTextWavePass.enable, True )

  dut.elaborate()

  # Setup the simulation
  dut.apply( DefaultPassGroup() )
  dut.sim_reset()
  # Test vector
  vector = [
    #  i        inlong       out
    b128(0x1000000000000+1),    b128(2),    b128(0x1000000000000+3),
    b128(0x1000000000000+0),    b128(2),    b128(0x1000000000000+2),
    b128(0x1000000000000+0),    b128(2),    b128(0x1000000000000+2),
    b128(0x1000000000000+1),    b128(-2),   b128(0x1000000000000+-1),
    b128(0x1000000000000+1),    b128(-42),  b128(0x1000000000000+-41),
    b128(0x1000000000000+1),    b128(-4),   b128(0x1000000000000+-3),
    b128(0x1000000000000+1),    b128(2),    b128(0x1000000000000+3),
    b128(0x1000000000000+0),    b128(2),    b128(0x1000000000000+2),
    b128(0x1000000000000+1),    b128(2),    b128(0x1000000000000+3),
    b128(0x1000000000000+0),    b128(-5),   b128(0x1000000000000+-5),
  ]

  # Begin simulation
  for i, inlong, out in zip(vector[0::3], vector[1::3], vector[2::3]):
    dut.i @= i
    dut.inlong @= inlong
    dut.sim_tick()
    assert dut.out == out

  #print
  f = io.StringIO()
  dut.print_textwave()
  with redirect_stdout(f):
    dut.print_textwave()
  out = f.getvalue()
  for i in dut.get_metadata( PrintTextWavePass.textwave_dict ):
    dot = i.find(".")
    sliced = i[dot+1:]
    if sliced != "reset" and sliced != "clk":
      assert i[dot+1:] in out

def test_bitstruct():

  @bitstruct
  class XX:
    x: Bits16
    y: Bits16

  class Toy( Component ):
    def construct( s ):
      # Interfaces
      s.i = InPort( XX )
      s.inlong = InPort( Bits16 )
      s.out = OutPort( XX )
      s.state = Wire(Bits1)
      @update
      def add_upblk():
        # This update block models the behavior of a 32-bit adder
        s.out.x @= s.i.x + s.inlong
        s.out.y @= s.i.y + s.inlong
        if s.out.x[3]:
          s.state @= s.state +b1(1)
        else:
          s.state @= b1(0)

  # Create a toy component and elaborate it
  dut = Toy()

  dut.set_metadata( PrintTextWavePass.enable, True )

  dut.elaborate()

  # Setup the simulation
  dut.apply( DefaultPassGroup() )

  dut.sim_reset()
  # Test vector
  vector = [
    #  i        inlong       out
    XX(1,1),    b16(2),    XX(3,3),
    XX(0,0),    b16(2),    XX(2,2),
    XX(0,0),    b16(2),    XX(2,2),
    XX(1,1),    b16(-2),   XX(-1,-1),
    XX(1,1),    b16(-42),  XX(-41,-41),
    XX(1,1),    b16(-4),   XX(-3,-3),
    XX(1,1),    b16(2),    XX(3,3),
    XX(0,0),    b16(2),    XX(2,2),
    XX(1,1),    b16(2),    XX(3,3),
    XX(0,0),    b16(-5),   XX(-5,-5),
  ]

  # Begin simulation
  for i, inlong, out in zip(vector[0::3], vector[1::3], vector[2::3]):
    dut.i @= i
    dut.inlong @= inlong
    dut.sim_tick()
    assert dut.out == out

  #print
  f = io.StringIO()
  dut.print_textwave()
  with redirect_stdout(f):
    dut.print_textwave()
  out = f.getvalue()
  for i in dut.get_metadata( PrintTextWavePass.textwave_dict ):
    dot = i.find(".")
    sliced = i[dot+1:]
    if sliced != "reset" and sliced != "clk":
      assert i[dot+1:] in out



#=========================================================================
# VerilogPlaceholderConfigs.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jul 28, 2019
"""Configuration of Verilog placeholders."""

import os
from copy import deepcopy

from pymtl3.passes.PassConfigs import Checker
from pymtl3.passes.PlaceholderConfigs import PlaceholderConfigs, expand

from .VerilogPlaceholderPass import VerilogPlaceholderPass


class VerilogPlaceholderConfigs( PlaceholderConfigs ):

  VerilogOptions = {
    # Parameters
    # Map the names of parameters to their values
    # If {} is provided, use the parameters inferred from `construct` instead
    "params" : {},

    # Port name mapping
    # Map PyMTL port names to external port names
    "port_map" : {},

    # Expects the name of the top component in external source files
    # "" to use name of the current component to be imported
    "top_module" : "",

    # Expects path of the file that contains the top module
    "src_file" : "",

    # -f
    # Expects the path to the flist file; "" to disable this option
    "v_flist" : "",

    # -v
    # Expects a list of paths to Verilog files; [] to disable this option
    "v_libs" : [],

    # -I ( alias of -y and +incdir+ )
    # Expects a list of include paths; [] to disable this option
    "v_include" : [],

    # The separator used for name mangling
    "separator" : '__',
  }

  VerilogCheckers = {
    ("params", "port_map") : Checker( lambda v: isinstance(v, dict), "expects a dict"),

    "top_module": Checker( lambda v: isinstance(v, str) and v, "expects a non-empty string"),

    "src_file": Checker( lambda v: isinstance(v, str) and (os.path.isfile(expand(v)) or not v),
                "src_file should be a path to a file or an empty string!" ),

    "v_flist": Checker( lambda v: isinstance(v, str) and os.path.isfile(expand(v)) or v == "",
                         "expects a path to a file" ),

    "v_libs": Checker( lambda v: isinstance(v, list) and all(os.path.exists(expand(p)) for p in v),
                            "expects a list of paths to files"),

    "v_include": Checker( lambda v: isinstance(v, list) and all(os.path.isdir(expand(p)) for p in v),
                            "expects a list of paths to directory"),
  }

  Pass = VerilogPlaceholderPass

  def __new__( cls, m ):
    inst = super().__new__( cls )

    # Do not pollute the attributes of the parent class
    cls.Options  = deepcopy( PlaceholderConfigs.Options )
    cls.Checkers = deepcopy( PlaceholderConfigs.Checkers )

    for key, val in cls.VerilogOptions.items():
      assert key not in cls.Options,\
        f'config {key} is duplicated between PlaceholderConfigs and VerilogPlaceholderConfigs'
      cls.Options[key] = val

    for cfgs, chk in cls.VerilogCheckers.items():
      if isinstance( cfgs, tuple ):
        for cfg in cfgs:
          inst._add_to_checkers( cls.Checkers, cfg, chk )
      elif isinstance( cfgs, str ):
        inst._add_to_checkers( cls.Checkers, cfgs, chk )

    return inst

  # Override
  def check( s ):
    super().check()

    # Exactly one of src_file and v_flist should be non-empty
    if (s.v_flist and os.path.isfile(expand(s.src_file))) or \
       (not s.v_flist and not os.path.isfile(expand(s.src_file))):
      raise InvalidPassOptionValue( 'src_file', s.src_file, s.Pass.__name__,
          'exactly one of src_file and v_flist should be non-emtpy!' )

  def get_port_map( s ):
    pmap = { p._dsl.full_name: name for p, name in s.port_map.items() }
    return lambda name: pmap[name] if name in pmap else name

  def _add_to_checkers( s, checkers, cfg, chk ):
    assert cfg not in checkers,\
      f'config {cfg} is duplicated between PlaceholderConfigs!'
    checkers[cfg] = chk

  # override to avoid deepcopying ports
  def collect_all_pass_configs( s, m ):
    c = s.__class__
    for opt, default in c.Options.items():
      assert not hasattr( s, opt ), f"There is already a field in config called '{opt}'. What happened?"
      if not m.has_metadata( getattr( c.Pass, opt ) ):
        setattr( s, opt, default )
      else:
        setattr( s, opt, m.get_metadata( getattr( c.Pass, opt ) ) )



"""
========================================================================
Tests for queues
========================================================================

Author: Shunning Jiang, Yanghui Ou
  Date: Feb 22, 2021
"""
from itertools import product

import pytest

from pymtl3 import *
from pymtl3.stdlib.test_utils import TestVectorSimulator, run_sim
from ..SourceRTL import SourceRTL
from ..SinkRTL import SinkRTL

from ..queues import (
    BypassQueue1EntryRTL,
    BypassQueueRTL,
    NormalQueue1EntryRTL,
    NormalQueueRTL,
    PipeQueue1EntryRTL,
    PipeQueueRTL,
)

#-------------------------------------------------------------------------
# TestVectorSimulator test
#-------------------------------------------------------------------------

def run_tv_test( dut, test_vectors, cmdline_opts  ):

  # Define input/output functions

  def tv_in( dut, tv ):
    dut.recv.val @= tv[0]
    dut.recv.msg @= tv[2]
    dut.send.rdy @= tv[4]

  def tv_out( dut, tv ):
    if tv[1] != '?': assert dut.recv.rdy == tv[1]
    if tv[3] != '?': assert dut.send.val == tv[3]
    if tv[5] != '?': assert dut.send.msg == tv[5]

  # Run the test

  sim = TestVectorSimulator( dut, test_vectors, tv_in, tv_out )
  sim.run_test( cmdline_opts )

def test_normal_behavior( cmdline_opts ):

  B1  = mk_bits(1)
  B32 = mk_bits(32)
  run_tv_test( NormalQueueRTL( Bits32, 2 ), [
    #  recv.val recv.rdy recv.msg  send.val send.rdy send.msg
    [  1,       1,       0x123,    0,       0,       '?'   ],
    [  1,       1,       0x345,    1,       0,       0x123 ],
    [  1,       0,       0x567,    1,       0,       0x123 ],
    [  1,       0,       0x567,    1,       1,       0x123 ],
    [  1,       1,       0x567,    1,       1,       0x345 ],
    [  1,       1,       0x890,    1,       0,       '?'   ],
    [  1,       0,       0x0  ,    1,       1,       0x567 ],
    [  0,       1,       0x0  ,    1,       1,       0x890 ],
    [  1,       1,       0x2  ,    0,       1,       '?'   ],
    [  1,       1,       0x3  ,    1,       1,       0x2   ],
    [  1,       1,       0x4  ,    1,       1,       0x3   ],
  ], cmdline_opts )

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, QueueType, src_msgs, sink_msgs ):

    s.src  = SourceRTL( MsgType, src_msgs )
    s.dut  = QueueType( MsgType )
    s.sink = SinkRTL( MsgType, sink_msgs )

    s.src.send //= s.dut.recv
    s.dut.send //= s.sink.recv

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return "{} ({}) {}".format(
      s.src.line_trace(), s.dut.line_trace(), s.sink.line_trace() )

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

test_msgs = [ Bits16( 4 ), Bits16( 1 ), Bits16( 2 ), Bits16( 3 ) ]

arrival_normal = [ 3, 5, 7, 9 ]
arrival_pipe   = [ 3, 4, 5, 6 ]
arrival_bypass = [ 2, 3, 4, 5 ]


def test_normal1_simple( cmdline_opts ):
  th = TestHarness( Bits16, NormalQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", arrival_time = arrival_normal )
  th.set_param( "top.dut.construct", num_entries = 1 )
  run_sim( th, cmdline_opts, duts=['dut'] )

def test_normal2_simple( cmdline_opts ):
  th = TestHarness( Bits16, NormalQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", arrival_time = arrival_pipe )
  th.set_param( "top.dut.construct", num_entries = 2 )
  run_sim( th, cmdline_opts, duts=['dut'] )

def test_normal3_simple( cmdline_opts ):
  th = TestHarness( Bits16, NormalQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", arrival_time = arrival_pipe )
  th.set_param( "top.dut.construct", num_entries = 3 )
  run_sim( th, cmdline_opts, duts=['dut'] )

def test_pipe1_simple( cmdline_opts ):
  th = TestHarness( Bits16, PipeQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", arrival_time = arrival_pipe )
  th.set_param( "top.dut.construct", num_entries = 1 )
  run_sim( th, cmdline_opts, duts=['dut'] )

def test_pipe1_backpressure( cmdline_opts ):
  th = TestHarness( Bits16, PipeQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", initial_delay = 20 )
  th.set_param( "top.dut.construct", num_entries = 1 )
  run_sim( th, cmdline_opts, duts=['dut'] )

def test_pipe2_backpressure( cmdline_opts ):
  th = TestHarness( Bits16, PipeQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", initial_delay = 20 )
  th.set_param( "top.dut.construct", num_entries = 2 )
  run_sim( th, cmdline_opts, duts=['dut'] )

def test_pipe3_backpressure( cmdline_opts ):
  th = TestHarness( Bits16, PipeQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", initial_delay = 20 )
  th.set_param( "top.dut.construct", num_entries = 3 )
  run_sim( th, cmdline_opts, duts=['dut'] )

def test_bypass1_simple( cmdline_opts ):
  th = TestHarness( Bits16, BypassQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", arrival_time = arrival_bypass )
  th.set_param( "top.dut.construct", num_entries = 1 )
  run_sim( th, cmdline_opts, duts=['dut'] )

def test_bypass1_backpressure( cmdline_opts ):
  th = TestHarness( Bits16, BypassQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", initial_delay = 20 )
  th.set_param( "top.dut.construct", num_entries = 1 )
  run_sim( th, cmdline_opts, duts=['dut'] )

def test_bypass2_sparse( cmdline_opts ):
  th = TestHarness( Bits16, BypassQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.src.construct", interval_delay = 3 )
  th.set_param( "top.dut.construct", num_entries = 2 )
  run_sim( th, cmdline_opts, duts=['dut'] )

def test_bypass3_sparse( cmdline_opts ):
  th = TestHarness( Bits16, BypassQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.src.construct", interval_delay = 3 )
  th.set_param( "top.dut.construct", num_entries = 3 )
  run_sim( th, cmdline_opts, duts=['dut'] )

@pytest.mark.parametrize(
  'QType, num_entries',
  product( [ NormalQueueRTL, PipeQueueRTL, BypassQueueRTL ], [ 8, 10, 12, 16 ] )
)
def test_large_backpressure( QType, num_entries, cmdline_opts ):
  msgs = test_msgs * 8
  th = TestHarness( Bits16, QType, msgs, msgs )
  th.set_param( "top.sink.construct", initial_delay = 20 )
  th.set_param( "top.dut.construct", num_entries = 16 )
  run_sim( th, cmdline_opts, duts=['dut'] )

@pytest.mark.parametrize(
  'QType', [ NormalQueue1EntryRTL, PipeQueue1EntryRTL, BypassQueue1EntryRTL ]
)
def test_single_simple( QType, cmdline_opts ):
  th = TestHarness( Bits16, QType, test_msgs, test_msgs )
  run_sim( th, cmdline_opts, duts=['dut'] )

@pytest.mark.parametrize(
  'QType', [ NormalQueue1EntryRTL, PipeQueue1EntryRTL, BypassQueue1EntryRTL ]
)
def test_single_backpressure( QType, cmdline_opts ):
  th = TestHarness( Bits16, QType, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", initial_delay = 10, interval_delay=2 )
  run_sim( th, cmdline_opts, duts=['dut'] )


@pytest.mark.parametrize(
  'QType, num_entries',
  product( [ NormalQueueRTL, PipeQueueRTL, BypassQueueRTL ], [ 8, 10, 12, 16 ] )
)
def test_large_delay( QType, num_entries, cmdline_opts ):
  msgs = test_msgs * 8
  th = TestHarness( Bits16, QType, msgs, msgs )
  th.set_param( "top.src.construct",  initial_delay=5,  interval_delay = 3)
  th.set_param( "top.sink.construct", initial_delay=20, interval_delay = 7)
  th.set_param( "top.dut.construct", num_entries = 16 )
  run_sim( th, cmdline_opts, duts=['dut'] )



from pymtl3 import *
from pymtl3.extra.pypy.fast_bytearray_funcs import (
    read_bytearray_bits,
    write_bytearray_bits,
)

from .mem_ifcs import MemMinionIfcFL
from .MemMsg import MemMsgType

AMO_FUNS = { MemMsgType.AMO_ADD  : lambda m,a : m+a,
             MemMsgType.AMO_AND  : lambda m,a : m&a,
             MemMsgType.AMO_OR   : lambda m,a : m|a,
             MemMsgType.AMO_SWAP : lambda m,a : a,
             MemMsgType.AMO_MIN  : lambda m,a : m if m.int() < a.int() else a,
             MemMsgType.AMO_MINU : min,
             MemMsgType.AMO_MAX  : lambda m,a : m if m.int() > a.int() else a,
             MemMsgType.AMO_MAXU : max,
             MemMsgType.AMO_XOR  : lambda m,a : m^a,
           }

class MagicMemoryFL( Component ):

  def construct( s, mem_nbytes=1<<20 ):
    s.mem = bytearray( mem_nbytes )

    s.ifc = MemMinionIfcFL( s.read, s.write, s.amo )

    s.trace = "     "
    @update_once
    def up_clear_trace():
      s.trace = "     "

  def read( s, addr, nbytes ):
    s.trace = "[rd ]"
    return read_bytearray_bits( s.mem, addr, nbytes )

  def write( s, addr, nbytes, data ):
    s.trace = "[wr ]"
    write_bytearray_bits( s.mem, addr, nbytes, data )

    # addr = int(addr)
    # end  = addr + nbytes

    # while addr < end:
      # s.mem[addr] = data & 255
      # data >>= 8
      # addr += 1
    # s.trace = "[wr ]"

  def amo( s, amo, addr, nbytes, data ):
    ret = s.read( addr, nbytes )
    s.write( addr, nbytes, AMO_FUNS[ int(amo) ]( ret, data ) )
    s.trace = "[amo]"
    return ret

  def read_mem( s, addr, size ):
    assert len(s.mem) > (addr + size)
    return s.mem[ addr : addr + size ]

  def write_mem( s, addr, data ):
    assert len(s.mem) > (addr + len(data))
    s.mem[ addr : addr + len(data) ] = data

  def line_trace( s ):
    return s.trace



#=========================================================================
# TranslationImport_adhoc_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 5, 2019
"""Test ad-hoc components with SystemVerilog translation and import."""

import gc
import resource

import pytest

from pymtl3 import *
from pymtl3.passes.backends.verilog import VerilogPlaceholderPass

from .. import VerilogTranslationImportPass


def _run_case():

  class DUT( Component ):
    def construct( s ):
      s.mem = bytearray(1<<29)
      s.mem[-1]=1
      s.mem[-2]=123
      s.mem[1] =12
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @update
      def upblk():
        s.out @= s.in_ + 1
    def line_trace( s ):
      return f"{s.in_} > {s.out}"

  m = DUT()
  m.elaborate()
  m.set_metadata( VerilogTranslationImportPass.enable, True )
  m.apply( VerilogPlaceholderPass() )
  m = VerilogTranslationImportPass()( m )

  m.apply( DefaultPassGroup() )
  m.sim_reset()

  m.in_ @= Bits32(123)
  m.sim_tick()
  assert m.out == 124

  m.finalize()
  gc.collect()
  print("python process memory usage:", resource.getrusage(resource.RUSAGE_SELF).ru_maxrss//1024," MB")

def test_verilog_translation_import_adhoc():
  for i in range(10):
    _run_case()



#=========================================================================
# BehavioralTranslatorL3.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 22, 2019
"""Provide L3 behavioral translator."""

from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL3Pass import (
    BehavioralRTLIRGenL3Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL3Pass import (
    BehavioralRTLIRTypeCheckL3Pass,
)

from .BehavioralTranslatorL2 import BehavioralTranslatorL2


class BehavioralTranslatorL3( BehavioralTranslatorL2 ):

  #-----------------------------------------------------------------------
  # _gen_behavioral_trans_metadata
  #-----------------------------------------------------------------------

  # Override
  def _gen_behavioral_trans_metadata( s, m ):
    m.apply( BehavioralRTLIRGenL3Pass( s.tr_top ) )
    m.apply( BehavioralRTLIRTypeCheckL3Pass( s.tr_top ) )
    s.behavioral.rtlir[m] = \
        m.get_metadata( BehavioralRTLIRGenL3Pass.rtlir_upblks )
    s.behavioral.freevars[m] =\
        m.get_metadata( BehavioralRTLIRTypeCheckL3Pass.rtlir_freevars )
    s.behavioral.tmpvars[m] =\
        m.get_metadata( BehavioralRTLIRTypeCheckL3Pass.rtlir_tmpvars )

  #-----------------------------------------------------------------------
  # Freevar datatype dispatch
  #-----------------------------------------------------------------------

  def dispatch_freevar_datatype( s, dtype ):
    if isinstance( dtype, rdt.Struct ):
      return s.rtlir_tr_struct_dtype( dtype )
    else:
      return super().dispatch_freevar_datatype( dtype )



#=========================================================================
# run_test_vector_sim_test
#=========================================================================

import pytest

from pymtl3 import *
from pymtl3.stdlib.test_utils import run_test_vector_sim
from pymtl3.stdlib.test_utils import RunTestVectorSimError

#-------------------------------------------------------------------------
# Test component with 1 input and 1 output port
#-------------------------------------------------------------------------

class TestComponent1i1o( Component ):
  def construct( s ):
    s.in_ = InPort(2)
    s.out = OutPort(2)

    @update
    def up():
      s.out @= s.in_ + 1

  def line_trace(s):
    return f"{s.in_}(){s.out}"

def test_1i1o_basic( cmdline_opts ):
  run_test_vector_sim( TestComponent1i1o(), [
    ('in_ out*' ),
    [ 0,  1     ],
    [ 1,  2     ],
    [ 0,  1     ],
  ], cmdline_opts )

# Verify that using a don't care works

def test_1i1o_dontcare( cmdline_opts ):
  run_test_vector_sim( TestComponent1i1o(), [
    ('in_ out*' ),
    [ 0,  '?'   ],
    [ 1,  2     ],
    [ 0,  1     ],
  ], cmdline_opts )

# Verify that an exception is thrown if the output is incorrect

def test_1i1o_incorrect_output( cmdline_opts ):
  with pytest.raises(RunTestVectorSimError):
    run_test_vector_sim( TestComponent1i1o(), [
      ('in_ out*' ),
      [ 0,  0     ],
      [ 1,  1     ],
      [ 0,  0     ],
    ], cmdline_opts )

# Verify that an exception is thrown if input port name is wrong

def test_1i1o_incorrect_inport( cmdline_opts ):
  with pytest.raises(RunTestVectorSimError):
    run_test_vector_sim( TestComponent1i1o(), [
      ('inX out*' ),
      [ 0,  0     ],
      [ 1,  2     ],
      [ 0,  0     ],
    ], cmdline_opts )

# Verify that an exception is thrown if output port name is wrong

def test_1i1o_incorrect_outport( cmdline_opts ):
  with pytest.raises(RunTestVectorSimError):
    run_test_vector_sim( TestComponent1i1o(), [
      ('in_ outX*' ),
      [ 0,  0      ],
      [ 1,  2      ],
      [ 0,  0      ],
    ], cmdline_opts )

#-------------------------------------------------------------------------
# Test component with 2 input and 2 output port
#-------------------------------------------------------------------------

class TestComponent2i2o( Component ):
  def construct( s ):
    s.in0  = InPort(2)
    s.in1  = InPort(2)
    s.out0 = OutPort(2)
    s.out1 = OutPort(2)

    @update
    def up():
      s.out0 @= s.in0 + 1
      s.out1 @= s.in1 + 2

  def line_trace(s):
    return f"{s.in0}|{s.in1}(){s.out0}|{s.out1}"

def test_2i2o_basic( cmdline_opts ):
  run_test_vector_sim( TestComponent2i2o(), [
    ('in0 in1 out0* out1*' ),
    [ 0,  0,  1,    2      ],
    [ 1,  0,  2,    2      ],
    [ 0,  1,  1,    3      ],
  ], cmdline_opts )

# Verify that using a don't care works

def test_2i2o_dontcare( cmdline_opts ):
  run_test_vector_sim( TestComponent2i2o(), [
    ('in0 in1 out0* out1*' ),
    [ 0,  0,  '?',  '?'    ],
    [ 1,  0,  2,    '?'    ],
    [ 0,  1,  1,    3      ],
  ], cmdline_opts )

# Verify that an exception is thrown if the output is incorrect

def test_2i2o_incorrect_output( cmdline_opts ):
  with pytest.raises(RunTestVectorSimError):
    run_test_vector_sim( TestComponent2i2o(), [
      ('in0 in1 out0* out1*' ),
      [ 0,  0,  1,    2      ],
      [ 1,  0,  3,    2      ],
      [ 0,  1,  1,    3      ],
    ], cmdline_opts )

# Verify that an exception is thrown if input port name is wrong

def test_2i2o_incorrect_inport( cmdline_opts ):
  with pytest.raises(RunTestVectorSimError):
    run_test_vector_sim( TestComponent2i2o(), [
      ('in0 inX out0* out1*' ),
      [ 0,  0,  1,    2      ],
      [ 1,  0,  2,    2      ],
      [ 0,  1,  1,    3      ],
    ], cmdline_opts )

# Verify that an exception is thrown if output port name is wrong

def test_2i2o_incorrect_outport( cmdline_opts ):
  with pytest.raises(RunTestVectorSimError):
    run_test_vector_sim( TestComponent2i2o(), [
      ('in0 in1 outX* out1*' ),
      [ 0,  0,  1,    2      ],
      [ 1,  0,  2,    2      ],
      [ 0,  1,  1,    3      ],
    ], cmdline_opts )

#-------------------------------------------------------------------------
# Test component with input and output array ports
#-------------------------------------------------------------------------

class TestComponentArray2i2o( Component ):
  def construct( s ):
    s.in_  = [ InPort(2)  for _ in range(2) ]
    s.out  = [ OutPort(2) for _ in range(2) ]

    @update
    def up():
      s.out[0] @= s.in_[0] + 1
      s.out[1] @= s.in_[1] + 2

  def line_trace(s):
    return f"{s.in_[0]}|{s.in_[1]}(){s.out[0]}|{s.out[1]}"

def test_a2i2o_basic( cmdline_opts ):
  run_test_vector_sim( TestComponentArray2i2o(), [
    ('in_[0] in_[1] out[0]* out[1]*' ),
    [ 0,     0,     1,      2        ],
    [ 1,     0,     2,      2        ],
    [ 0,     1,     1,      3        ],
  ], cmdline_opts )

# Verify that using a don't care works

def test_a2i2o_dontcare( cmdline_opts ):
  run_test_vector_sim( TestComponentArray2i2o(), [
    ('in_[0] in_[1] out[0]* out[1]*' ),
    [ 0,     0,     '?',    '?'      ],
    [ 1,     0,     2,      '?'      ],
    [ 0,     1,     1,      3        ],
  ], cmdline_opts )

# Verify that an exception is thrown if the output is incorrect

def test_a2i2o_incorrect_output( cmdline_opts ):
  with pytest.raises(RunTestVectorSimError):
    run_test_vector_sim( TestComponentArray2i2o(), [
      ('in_[0] in_[1] out[0]* out[1]*' ),
      [ 0,     0,     1,      2        ],
      [ 1,     0,     3,      2        ],
      [ 0,     1,     1,      3        ],
    ], cmdline_opts )

# Verify that an exception is thrown if input port name is wrong

def test_a2i2o_incorrect_inport( cmdline_opts ):
  with pytest.raises(RunTestVectorSimError):
    run_test_vector_sim( TestComponentArray2i2o(), [
      ('in_[0] inX    out[0]* out[1]*' ),
      [ 0,     0,     1,      2        ],
      [ 1,     0,     2,      2        ],
      [ 0,     1,     1,      3        ],
    ], cmdline_opts )

# Verify that an exception is thrown if output port name is wrong

def test_a2i2o_incorrect_outport( cmdline_opts ):
  with pytest.raises(RunTestVectorSimError):
    run_test_vector_sim( TestComponentArray2i2o(), [
      ('in_[0] in_[1] outX*   out[1]*' ),
      [ 0,     0,     1,      2        ],
      [ 1,     0,     2,      2        ],
      [ 0,     1,     1,      3        ],
    ], cmdline_opts )




"""
========================================================================
helpers.py
========================================================================
Useful helper functions of Bits operations. Ideally these functions
should be implemented in RPython.

Author : Shunning Jiang
Date   : Nov 3, 2017
"""
import math

from .bits_import import *

try:
  from mamba import concat
except:
  def concat( *args ):
    value = nbits = 0

    for x in args:
      xnb = x.nbits
      nbits += xnb
      value = (value << xnb) | x.uint()

    return Bits( nbits, value )

def trunc( value, new_width ):
  if isinstance( new_width, int ):
    assert new_width <= value.nbits
    return Bits( new_width, value.uint(), trunc_int=True )
  else:
    assert issubclass( new_width, Bits )
    return new_width( value.uint(), trunc_int=True )

def zext( value, new_width ):
  if isinstance( new_width, int ):
    assert new_width >= value.nbits
    return Bits( new_width, value.uint() )
  else:
    assert issubclass( new_width, Bits )
    return new_width( value.uint() )

def clog2( N ):
  assert N > 0
  return int( math.ceil( math.log( N, 2 ) ) )

def sext( value, new_width ):
  if isinstance( new_width, int ):
    assert new_width >= value.nbits
    return Bits( new_width, value.int() )
  else:
    assert issubclass( new_width, Bits )
    return new_width( value.int() )

def reduce_and( value ):
  try:
    return b1( int(value) == (1 << value.nbits) - 1 )
  except AttributeError:
    raise TypeError("Cannot call reduce_and on int")

def reduce_or( value ):
  try:
    return b1( int(value) != 0 )
  except AttributeError:
    raise TypeError("Cannot call reduce_or on int")

def reduce_xor( value ):
  try:
    pop_count = 0
    value = int(value)
    while value != 0:
      pop_count += value & 1
      value >>= 1
    return b1( pop_count & 1 )

  except AttributeError:
    raise TypeError("Cannot call reduce_xor on int")



#=========================================================================
# StructuralRTLIRGenL4Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Apr 3, 2019
"""Provide L4 structural RTLIR generation pass."""

from .StructuralRTLIRGenL3Pass import StructuralRTLIRGenL3Pass


class StructuralRTLIRGenL4Pass( StructuralRTLIRGenL3Pass ):
  """At L4 we need to recursively generate metadata for every component"""

  # Override
  def _gen_metadata( s, m ):
    super()._gen_metadata( m )
    for child in m.get_child_components(repr):
      s._gen_metadata( child )



"""
========================================================================
PrintWavePass.py
========================================================================
Print single bit signal in wave form and multi-bit signal by showing
least significant bits.

To use, call top.print_wave()

Inspired by PyRTL's state machine screenshot, which shows the change of signal
values along ticks of the clock.

Link: https://ucsbarchlab.github.io/PyRTL/

Author : Kaishuo Cheng, Shunning Jiang
Date   : Nov 9, 2019
"""

import py

from pymtl3.dsl import Const, MetadataKey
from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.errors import PassOrderError


class PrintTextWavePass( BasePass ):

  # PrintWavePass public pass data

  #: chars_per_cycle
  #:
  #: Type: ``str``; input
  #:
  #: Default value: ""
  chars_per_cycle = MetadataKey(int)

  #: enable
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: False
  enable = MetadataKey(bool)

  textwave_func = MetadataKey()
  textwave_dict = MetadataKey()

  def __call__( self, top ):
    if top.has_metadata( self.enable ) and top.get_metadata( self.enable ):

      assert not top.has_metadata( self.textwave_func )
      assert not top.has_metadata( self.textwave_dict )

      func, sigs_dict = self._collect_sig_func( top )

      top.set_metadata( self.textwave_func, func )
      top.set_metadata( self.textwave_dict, sigs_dict )
      top.print_textwave = self._gen_print_wave( top, sigs_dict )

  def _process_binary( self, sig, base, max ):
    """
    Returns int value from a signal in 32b form. Used for testing.

    Example: input: 0b00000000000000000000000000000000 10
             output: 0
    """
    if sig[1] == "b":
      sig = sig[2:]
    if base == 10:
      temp_int = int(sig,2)
      if sig[0] == '1':
        return temp_int -2 **32     # taking 2's complement.
                                    # leading 1 indicates a negative number
      else:
        return temp_int
      #hex number
    else:
      temp_hex = hex(int(sig,2))[2:]
      l = len(temp_hex)
      if l > max:
        temp_hex = temp_hex[l-max:]

      if l < max:
        temp_hex = '0'*(max-l) + temp_hex
      return temp_hex

  def _gen_print_wave( self, top, sigs_dict ):

    def print_wave():
      if top.has_metadata( self.chars_per_cycle ):
        char_length = top.get_metadata( self.chars_per_cycle )
      else:
        char_length = 6
      # Shunning: deprecate text_fancy
      # up, down = '\u2571', '\u2572'
      # x, low = '\u2573', '\u005f'

      assert char_length % 2 == 0
      tick = '|'
      up,down,x,low,high = '/','\\','|','_', '\u203e'
      revstart, revstop = '\x1B[7m', '\x1B[0m'
      light_gray = '\033[47m'
      back='\033[0m'  #back to normal printing

      all_signal_values = text_sigs
      #spaces before cycle number
      max_length = 5
      for sig in all_signal_values:
        #   Example: s.in(12b)
        # length of signal name + (b) + number of digits, like 12
        # to add(32b) in front, add this to maxlength:
        #len(str(len(all_signal_values[sig][0][0])))+3
        max_length = max( max_length, len(sig)-2 )

      print("")
      print(" "*(max_length+1),end = "")

      #-----------------------------------------------------------------------
      # handle clk and reset
      #-----------------------------------------------------------------------
      # handles clock tick symbol

      for i in range(len(all_signal_values["s.reset"])):
        # insert a space every 5 cycles
        print(f"{tick}{str(i).ljust(char_length-1)}",end="")
      print("")

      # Adding one blank line
      print("")

      # handle clock signal
      clk_cycle_str = up + (char_length-2)//2*str(high) + down + (char_length-2)//2*str(low)

      print("clk".rjust(max_length), clk_cycle_str * len(all_signal_values["s.reset"]))

      print("")

      #signals
      for sig in all_signal_values:
        bit_length = len(all_signal_values[sig][0])-2
        print((sig[2:]).rjust(max_length),end=" ")
        # one bit
        if bit_length==1:
          prev_sig = None
          for i, val in enumerate( all_signal_values[sig] ):
            # every 5 cycles add a space
            # if i%5 == 0:
              # print(" ",end = "")
            if val[2] == '1':
              current_sig = high
            else:
              current_sig = low
            # detecting if first cycle
            if prev_sig is not None:
              if prev_sig == low and current_sig == high:
                print(up+current_sig*(char_length-1),end = "")
              elif prev_sig == high and current_sig == low:
                print(down+current_sig*(char_length-1),end = "")
              # prev and current signal agree
              else:
                print(current_sig*char_length,end = "")
            # first cycle
            else:
              print(current_sig*char_length,end = "")
            prev_sig = current_sig

          print("")
          #multiple bits
        else:
          next = 0
          val_list = all_signal_values[sig]
          for i in range(len(val_list)):
            # signals in this cycle is still the same as before
            if next > 0:
              next -= 1
              continue

            val = val_list[i]
            for j in range(i,len(val_list)):
              if val_list[j] != val:
                j = j-1
                break

            #first is reserved for X or " ". Rest is 5 char length.
            length = (char_length-1) + char_length*(j-i)
            next = j-i

            if length >= bit_length // (char_length-1):
              length = bit_length // (char_length-1)
              if bit_length % char_length != 0:
                length += 1
              plus = False
            else:
              #reverse a place for +
              length = length -1
              plus = True

            current = self._process_binary(val,16,length)
            # print a +, with one less space for signal number
            if plus:
              if i==0:
                print(light_gray + " " +'\033[30m'+"+"+ current,end = "")
              else:
                print(light_gray + '\033[30m'+x + "+" +current,end = "")
            # no +, more spaces for signal number
            else:
              if i==0:
                print(light_gray + " " +'\033[30m'+ current,end = "")
              else:
                print(light_gray + '\033[30m'+x +current,end = "")
              print(" "*(char_length-1+char_length*(j-i)-length),end = "")
          print(back + "")
        print("")
    return print_wave

  def _collect_sig_func( self, top ):

    # TODO use actual nets to reduce the amount of saved signals

    # Give all ' and " characters a preceding backslash for .format
    wav_srcs = []
    text_sigs = {}

    # Now we create per-cycle signal value collect functions
    signal_names = []
    for x in top._dsl.all_signals:
      if x.is_top_level_signal() and x.get_field_name() != "clk" and x.get_field_name() != "reset":
        signal_names.append( (x._dsl.level, repr(x)) )

    for _, x in [(0, 's.reset')] + sorted(signal_names):
      text_sigs[x] = []
      wav_srcs.append(f"text_sigs['{x}'].append( {x}.to_bits().bin() )")

    # TODO use integer index instead of dict, should be easy
    src =  """
def dump_wav():
  {}
""".format( "\n  ".join(wav_srcs) )
    s, l_dict = top, {}
    exec(compile( src, filename="temp", mode="exec"), globals().update(locals()), l_dict)
    return l_dict['dump_wav'], text_sigs
